# decompilation options

from collections import defaultdict


class DecompilationOption:
    def __init__(self, name, description, value_type, cls, param, value_range=None, category="General",
                 default_value=None):
        self.name = name
        self.description = description
        self.value_type = value_type
        self.cls = cls
        self.param = param
        self.value_range = value_range
        self.category = category
        self.default_value = default_value


O = DecompilationOption

options = [
    O(
        "Remove dead memdefs",
        "Allow the decompiler to remove memory definitions (such as stack variables) that are deemed dead. Generally, "
        "enabling this option will generate cleaner pseudocode; But when prior analyses go wrong, angr may miss "
        "certain uses to a memory definition, which may cause the removal of a memory definition that is in use.",
        bool,
        "clinic",
        "remove_dead_memdefs",
        category="Data flows",
        default_value=True,
    ),
    O(
        "Display exception edges (experimental)",
        "Decompile and display exception handling code. Enabling this option generally degrades the readability of the "
        "pseudo code. This is still an experimental feature.",
        bool,
        "clinic",
        "exception_edges",
        category="Graph",
        default_value=False,
    )
]

options_by_category = defaultdict(list)

for o in options:
    options_by_category[o.category].append(o)



import copy
from typing import Dict, Tuple, Any

from claripy.vsa import AbstractLocation

from .. import MemoryMixin


class MemoryRegionMetaMixin(MemoryMixin):

    __slots__ = ('_endness', '_id', '_state', '_is_stack', '_related_function_addr', '_alocs', '_memory', )

    def __init__(self, related_function_addr=None, **kwargs):
        super().__init__(**kwargs)
        self._related_function_addr = related_function_addr
        # This is a map from tuple (basicblock_key, stmt_id) to AbstractLocation objects
        self.alocs: Dict[Tuple[Any,int],AbstractLocation] = { }

        self._is_stack = None

    @property
    def is_stack(self):
        if self.id is None:
            return None
        if self._is_stack is None:
            self._is_stack = self.id.startswith('stack_')
        return self._is_stack

    @property
    def related_function_addr(self):
        return self._related_function_addr

    def get_abstract_locations(self, addr, size):
        """
        Get a list of abstract locations that is within the range of [addr, addr + size]

        This implementation is pretty slow. But since this method won't be called frequently, we can live with the bad
        implementation for now.

        :param addr:    Starting address of the memory region.
        :param size:    Size of the memory region, in bytes.
        :return:        A list of covered AbstractLocation objects, or an empty list if there is none.
        """

        ret = [ ]
        for aloc in self.alocs.values():
            for seg in aloc.segments:
                if seg.offset >= addr and seg.offset < addr + size:
                    ret.append(aloc)
                    break

        return ret

    @MemoryMixin.memo
    def copy(self, memo):
        r: 'MemoryRegionMetaMixin' = super().copy(memo)
        r._alocs = copy.deepcopy(self.alocs)
        r._related_function_addr = self._related_function_addr
        return r

    def store(self, addr, data, bbl_addr=None, stmt_id=None, ins_addr=None, endness=None, **kwargs):
        if ins_addr is not None:
            aloc_id = ins_addr
        else:
            # It comes from a SimProcedure. We'll use bbl_addr as the aloc_id
            aloc_id = bbl_addr

        if aloc_id not in self.alocs:
            self.alocs[aloc_id] = self.state.solver.AbstractLocation(bbl_addr,
                                                                      stmt_id,
                                                                      self.id,
                                                                      region_offset=addr,
                                                                      size=len(data) // self.state.arch.byte_width)
            return super().store(addr, data, endness=endness, **kwargs)
        else:
            if self.alocs[aloc_id].update(addr, len(data) // self.state.arch.byte_width):
                return super().store(addr, data, endness=endness, **kwargs)
            else:
                return super().store(addr, data, endness=endness, **kwargs)

    def load(self, addr, size=None, bbl_addr=None, stmt_idx=None, ins_addr=None, **kwargs): #pylint:disable=unused-argument
        #if bbl_addr is not None and stmt_id is not None:
        return super().load(addr, size=size, **kwargs)

    def _merge_alocs(self, other_region):
        """
        Helper function for merging.
        """
        merging_occurred = False
        for aloc_id, aloc in other_region.alocs.items():
            if aloc_id not in self.alocs:
                self.alocs[aloc_id] = aloc.copy()
                merging_occurred = True
            else:
                # Update it
                merging_occurred |= self.alocs[aloc_id].merge(aloc)
        return merging_occurred

    def merge(self, others, merge_conditions, common_ancestor=None) -> bool:
        r = False
        for other_region in others:
            self._merge_alocs(other_region)
            r |= super().merge(
                [other_region], merge_conditions, common_ancestor=common_ancestor
            )
        return r

    def widen(self, others):
        for other_region in others:
            self._merge_alocs(other_region)
            super().widen([ other_region.memory ])

    def dbg_print(self, indent=0):
        """
        Print out debugging information
        """
        print("%sA-locs:" % (" " * indent))
        for aloc_id, aloc in self.alocs.items():
            print("%s<0x%x> %s" % (" " * (indent + 2), aloc_id, aloc))



import claripy

from . import MemoryMixin
from ...errors import SimSegfaultException

class SmartFindMixin(MemoryMixin):
    def find(self, addr, data, max_search,
             default=None,
             endness=None,
             chunk_size=None,
             max_symbolic_bytes=None,
             condition=None,
             **kwargs):

        if endness is None:
            endness = self.endness
            if endness is None:
                endness = 'Iend_BE'

        stride = self._find_stride(data)

        if chunk_size is None:
            chunk_size = min(max_search, 0x80)

        cases = []
        match_indices = []
        constraints = []

        try:
            for i, (subaddr, element) in enumerate(self._find_iter_items(addr, stride, chunk_size, max_search, endness, condition, max_symbolic_bytes, **kwargs)):
                comparison, concrete_comparison = self._find_compare(element, data, **kwargs)

                if concrete_comparison is False:
                    continue

                match_indices.append(i*stride)
                if isinstance(subaddr, int):
                    subaddr = claripy.BVV(subaddr, size=self.state.arch.bits)
                cases.append((comparison, subaddr))

                if concrete_comparison is True:
                    break

            else:
                # the loop terminated, meaning we exhausted some sort of limit instead of finding a concrete answer.
                if default is None:
                    constraints.append(claripy.Or(*(c for c, _ in cases)))
        except SimSegfaultException:
            if chunk_size > 1:
                return self.find(addr, data, max_search,
                                 default=default,
                                 endness=endness,
                                 chunk_size=1,
                                 max_symbolic_bytes=max_symbolic_bytes,
                                 condition=condition,
                                 **kwargs)
            raise

        if len(cases) == 1:
            return cases[0][1], constraints, match_indices

        return self._find_process_cases(cases, match_indices, constraints, default, **kwargs)

    def _find_stride(self, target):
        if type(target) is bytes:
            return len(target)
        return len(target) // self.state.arch.byte_width

    def _find_iter_items(self, addr, stride, chunk_size, max_search, endness, condition, max_symbolic_bytes, **kwargs):
        if condition is None:
            condition = self.state.solver.true
        chunk = None
        chunk_progress = chunk_size
        for i in range(0, max_search, stride):
            subaddr = addr + i
            if chunk_progress == chunk_size:
                chunk = self.load(subaddr, size=stride*chunk_size, endness=endness, condition=condition & self._find_condition(addr + i, **kwargs), **kwargs)
                chunk_progress = 0

            chunk_idx = (chunk_progress if endness == 'Iend_BE' else chunk_size - 1 - chunk_progress)*stride
            chunk_progress += 1
            b = chunk.get_bytes(chunk_idx, stride)

            if self._find_are_bytes_symbolic(b) and max_symbolic_bytes is not None:
                if max_symbolic_bytes:
                    max_symbolic_bytes -= 1
                else:
                    return
            yield subaddr, b

    def _find_are_bytes_symbolic(self, b):
        return b.symbolic

    def _find_condition(self, target_addr, **kwargs):
        # TODO: fill this in in order to make each load have the correct condition associated with it
        return claripy.true

    def _find_compare(self, element, target, **kwargs):
        comparison = element == target
        if self.state.solver.is_true(comparison):
            concrete_comparison = True
        elif self.state.solver.is_false(comparison):
            concrete_comparison = False
        else:
            concrete_comparison = None
        return comparison, concrete_comparison

    def _find_process_cases(self, cases, match_indices, constraints, default, **kwargs):
        if default is None:
            default = claripy.BVV(0, self.state.arch.bits)
        if cases and cases[-1][0].is_true():
            default = cases.pop(-1)[1]
        result = self.state.solver.ite_cases(cases, default)
        return result, constraints, match_indices



import angr

######################################
# getsockopt
######################################

class getsockopt(angr.SimProcedure):
    #pylint:disable=arguments-differ

    def run(self, sockfd, level, optname, optval, optlen):

        # TODO: ...

        return 0



import angr

######################################
# fcntl
######################################

class fcntl(angr.SimProcedure):
    #pylint:disable=arguments-differ

    def run(self, fd, cmd):
        #  this is a stupid stub that does not do anything besides returning an unconstrained variable.
        return self.state.solver.BVS('fcntl_retval', self.state.arch.bits)



# pylint:disable=unused-import
from ..models import DbKnowledgeBase, DbXRefs
from ...knowledge_plugins.xrefs import XRefManager
from ...knowledge_plugins.cfg import CFGModel
from ...knowledge_base import KnowledgeBase


class XRefsSerializer:
    """
    Serialize/unserialize an XRefs object to/from a database session.
    """

    @staticmethod
    def dump(session, db_kb, xrefs):
        """

        :param session:
        :param DbKnowledgeBase db_kb:
        :param XRefManager xrefs:
        :return:
        """

        db_xrefs = db_kb.xrefs

        blob = xrefs.serialize()
        if db_xrefs is not None:
            # update the existing xrefs
            db_xrefs.blob = blob
        else:
            # create a new xrefs
            db_xrefs = DbXRefs(kb=db_kb, blob=blob)
            session.add(db_xrefs)

    @staticmethod
    def load(session, db_kb, kb, cfg_model=None):  # pylint:disable=unused-argument
        """

        :param session:
        :param DbKnowledgeBase db_kb:
        :param KnowledgeBase kb:
        :param CFGModel cfg_model:
        :return:
        """

        db_xrefs = db_kb.xrefs
        if db_xrefs is None:
            return None

        xrefs = XRefManager.parse(db_xrefs.blob, cfg_model=cfg_model, kb=kb)
        return xrefs



import logging

import networkx
from . import Analysis

l = logging.getLogger(name=__name__)

class Loop(object):
    def __init__(self, entry, entry_edges, break_edges, continue_edges, body_nodes, graph, subloops):
        self.entry = entry
        self.entry_edges = entry_edges
        self.break_edges = break_edges
        self.continue_edges = continue_edges
        self.body_nodes = body_nodes
        self.graph = graph
        self.subloops = subloops

        self.has_calls = any(map(lambda loop: loop.has_calls, subloops))

        if not self.has_calls:
            for _, _, data in self.graph.edges(data=True):
                if 'type' in data and data['type'] == 'fake_return':
                    # this is a function call.
                    self.has_calls = True
                    break

    def __repr__(self):
        s = "<Loop @ %s, %d blocks>" % (self.entry.addr, len(self.body_nodes))
        return s

class LoopFinder(Analysis):
    """
    Extracts all the loops from all the functions in a binary.
    """

    def __init__(self, functions=None, normalize=True):
        if functions is None:
            functions = self.kb.functions.values()

        found_any = False
        self.loops = []
        self.loops_hierarchy = {}
        for function in functions:

            if self.project.is_hooked(function.addr) or \
                    self.project.simos.is_syscall_addr(function.addr):
                # skip SimProcedures and syscalls
                continue

            found_any = True
            with self._resilience():
                if normalize:
                    function.normalize()
                tops, alls = self._parse_loops_from_graph(function.graph)
                self.loops += alls
                self.loops_hierarchy[function.addr] = tops

        if not found_any:
            l.error("No knowledge of functions is present. Did you forget to construct a CFG?")

    def _parse_loop_graph(self, subg, bigg):
        """
        Create a Loop object for a strongly connected graph, and any strongly
        connected subgraphs, if possible.

        :param subg:    A strongly connected subgraph.
        :param bigg:    The graph which subg is a subgraph of.

        :return:        A list of Loop objects, some of which may be inside others,
                        but all need to be documented.
        """
        loop_body_nodes = list(subg.nodes())[:]
        entry_edges = []
        break_edges = []
        continue_edges = []
        entry_node = None
        for node in loop_body_nodes:
            for pred_node in bigg.predecessors(node):
                if pred_node not in loop_body_nodes:
                    if entry_node is not None and entry_node != node:
                        l.warning("Bad loop: more than one entry point (%s, %s)", entry_node, node)
                        return None, []
                    entry_node = node
                    entry_edges.append((pred_node, node))
                    subg.add_edge(pred_node, node)
            for succ_node in bigg.successors(node):
                if succ_node not in loop_body_nodes:
                    break_edges.append((node, succ_node))
                    subg.add_edge(node, succ_node)
        if entry_node is None:
            entry_node = min(loop_body_nodes, key=lambda n: n.addr)
            l.info("Couldn't find entry point, assuming it's the first by address (%s)", entry_node)

        acyclic_subg = subg.copy()
        for pred_node in subg.predecessors(entry_node):
            if pred_node in loop_body_nodes:
                continue_edge = (pred_node, entry_node)
                acyclic_subg.remove_edge(*continue_edge)
                continue_edges.append(continue_edge)

        removed_exits = {}
        removed_entries = {}
        tops, alls = self._parse_loops_from_graph(acyclic_subg)
        for subloop in tops:
            if subloop.entry in loop_body_nodes:
                # break existing entry edges, exit edges
                # re-link in loop object
                # the exception logic is to handle when you have two loops adjacent to each other
                # you gotta link the two loops together and remove the dangling edge
                for entry_edge in subloop.entry_edges:
                    try:
                        subg.remove_edge(*entry_edge)
                    except networkx.NetworkXError:
                        if entry_edge in removed_entries:
                            subg.add_edge(removed_entries[entry_edge], subloop)
                            try:
                                subg.remove_edge(removed_entries[entry_edge], entry_edge[1])
                            except networkx.NetworkXError:
                                pass
                        else:
                            raise
                    else:
                        subg.add_edge(entry_edge[0], subloop)
                        removed_entries[entry_edge] = subloop
                for exit_edge in subloop.break_edges:
                    try:
                        subg.remove_edge(*exit_edge)
                    except networkx.NetworkXError:
                        if exit_edge in removed_entries:
                            subg.add_edge(subloop, removed_entries[exit_edge])
                            try:
                                subg.remove_edge(exit_edge[0], removed_entries[exit_edge])
                            except networkx.NetworkXError:
                                pass
                        else:
                            raise
                    else:
                        subg.add_edge(subloop, exit_edge[1])
                        removed_exits[exit_edge] = subloop
                _subgraphs = (networkx.induced_subgraph(subg, nodes).copy() for nodes in
                             networkx.weakly_connected_components(subg))
                subg = next(filter( lambda g: entry_node in g.nodes(),
                                    _subgraphs))
        me = Loop(entry_node,
             entry_edges,
             break_edges,
             continue_edges,
             loop_body_nodes,
             subg,
             tops[:])
        return me, [me] + alls

    def _parse_loops_from_graph(self, graph):
        """
        Return all Loop instances that can be extracted from a graph.

        :param graph:   The graph to analyze.

        :return:        A list of all the Loop instances that were found in the graph.
        """
        outtop = []
        outall = []
        for subg in ( networkx.induced_subgraph(graph, nodes).copy() for nodes in networkx.strongly_connected_components(graph)):
            if len(subg.nodes()) == 1:
                if len(list(subg.successors(list(subg.nodes())[0]))) == 0:
                    continue
            thisloop, allloops = self._parse_loop_graph(subg, graph)
            if thisloop is not None:
                outall += allloops
                outtop.append(thisloop)
        return outtop, outall

from angr.analyses import AnalysesHub
AnalysesHub.register_default('LoopFinder', LoopFinder)




import logging

from ...errors import AngrCallableError, AngrCallableMultistateError
from ...calling_conventions import DEFAULT_CC


l = logging.getLogger(name=__name__)
# l.setLevel("DEBUG")


class IdentifierCallable(object):
    """
    Callable is a representation of a function in the binary that can be
    interacted with like a native python function.

    If you set perform_merge=True (the default), the result will be returned to you, and
    you can get the result state with callable.result_state.

    Otherwise, you can get the resulting path group (immutable) at callable.result_path_group.
    """

    def __init__(self, project, addr, concrete_only=False, perform_merge=False, base_state=None, toc=None, cc=None,
                 max_steps=None):
        """
        :param project:         The project to operate on
        :param addr:            The address of the function to use

        The following parameters are optional:

        :param concrete_only:   Throw an exception if the execution splits into multiple paths
        :param perform_merge:   Merge all result states into one at the end (only relevant if concrete_only=False)
        :param base_state:      The state from which to do these runs
        :param toc:             The address of the table of contents for ppc64
        :param cc:              The SimCC to use for a calling convention
        """

        self._project = project
        self._addr = addr
        self._concrete_only = concrete_only
        self._perform_merge = perform_merge
        self._base_state = base_state
        self._toc = toc
        self._caller = None
        self._cc = cc if cc is not None else DEFAULT_CC[project.arch.name](project.arch)
        self._deadend_addr = project.simos.return_deadend
        self._max_steps = max_steps

        self.result_path_group = None
        self.result_state = None

    def set_base_state(self, state):
        """
        Swap out the state you'd like to use to perform the call
        :param state: The state to use to perform the call
        """
        self._base_state = state

    def __call__(self, *args):
        self.perform_call(*args)
        if self.result_state is not None:
            return self.result_state.solver.simplify(self._cc.get_return_val(self.result_state, stack_base=self.result_state.regs.sp - self._cc.STACKARG_SP_DIFF))
        return None

    def get_base_state(self, *args):
        self._base_state.ip = self._addr
        state = self._project.factory.call_state(self._addr, *args,
                    cc=self._cc,
                    base_state=self._base_state,
                    ret_addr=self._deadend_addr,
                    toc=self._toc)
        return state

    def perform_call(self, *args):
        self._base_state.ip = self._addr
        state = self._project.factory.call_state(self._addr, *args,
                    cc=self._cc,
                    base_state=self._base_state,
                    ret_addr=self._deadend_addr,
                    toc=self._toc)

        def step_func(pg):
            pg2 = pg.prune()
            if len(pg2.active) > 1:
                raise AngrCallableMultistateError("Execution split on symbolic condition!")
            return pg2

        caller = self._project.factory.simulation_manager(state)
        for _ in range(self._max_steps):
            if len(caller.active) == 0:
                break
            if caller.active[0].history.block_count > 100000:
                l.debug("super long path %s", caller.active[0])
                raise AngrCallableError("Super long path")
            caller.step(step_func=step_func if self._concrete_only else None)
        if len(caller.active) > 0:
            raise AngrCallableError("didn't make it to the end of the function")

        caller.unstash(from_stash='deadended')
        caller.prune(filter_func=lambda pt: pt.addr == self._deadend_addr)

        if len(caller.active) == 0:
            raise AngrCallableError("No paths returned from function")

        self.result_path_group = caller.copy()

        if self._perform_merge:
            caller.merge()
        self.result_state = caller.active[0]



from typing import Set, Dict, List
from collections import defaultdict

from ailment.block import Block
from ailment.statement import Statement, Assignment, Store, Call
from ailment.expression import Register, Convert, Load, StackBaseOffset

from ...code_location import CodeLocation
from ...analyses.reaching_definitions.external_codeloc import ExternalCodeLocation
from ...sim_variable import SimStackVariable
from ...analyses.propagator.propagator import Equivalence
from ...knowledge_plugins.key_definitions import atoms
from ...knowledge_plugins.key_definitions.definition import Definition
from .. import Analysis, AnalysesHub
from .ailblock_walker import AILBlockWalker


class HasCallNotification(Exception):
    pass


class AILSimplifier(Analysis):
    """
    Perform function-level simplifications.
    """
    def __init__(self, func, func_graph=None, remove_dead_memdefs=False, unify_variables=False,
                 reaching_definitions=None):
        self.func = func
        self.func_graph = func_graph if func_graph is not None else func.graph
        self._reaching_definitions = reaching_definitions

        self._remove_dead_memdefs = remove_dead_memdefs
        self._unify_vars = unify_variables

        self._calls_to_remove: Set[CodeLocation] = set()
        self.blocks = {}  # Mapping nodes to simplified blocks

        self._simplify()

    def _simplify(self):

        if self._unify_vars:
            self._unify_local_variables()
            self._fold_call_exprs()
        self._remove_dead_assignments()

    def _unify_local_variables(self):
        """
        Find variables that are definitely equivalent and then eliminate the unnecessary copies.
        """
        prop = self.project.analyses.Propagator(func=self.func, func_graph=self.func_graph)
        if not prop.equivalence:
            return

        addr2block: Dict[int, Block] = { }
        for block in self.func_graph.nodes():
            addr2block[block.addr] = block

        for eq in prop.equivalence:
            eq: Equivalence

            # Acceptable equivalence classes:
            #
            # stack variable == register
            # register variable == register
            # stack variable == Conv(register, M->N)
            #
            the_def = None
            if isinstance(eq.atom0, SimStackVariable):
                if isinstance(eq.atom1, Register):
                    # stack_var == register
                    reg = eq.atom1
                elif isinstance(eq.atom1, Convert) and isinstance(eq.atom1.operand, Register):
                    # stack_var == Conv(register, M->N)
                    reg = eq.atom1.operand
                else:
                    continue

            elif isinstance(eq.atom0, Register):
                if isinstance(eq.atom1, Register):
                    # register == register
                    reg = eq.atom1
                else:
                    continue

            else:
                continue

            # find the definition of this register
            defs = self._reaching_definitions.all_uses.get_uses_by_location(eq.codeloc)
            for def_ in defs:
                def_: Definition
                if isinstance(def_.atom, atoms.Register) and def_.atom.reg_offset == reg.reg_offset:
                    # found it!
                    the_def = def_
                    break

            if the_def is None:
                continue
            if isinstance(the_def.codeloc, ExternalCodeLocation):
                continue

            # find all uses of this definition
            all_uses: Set[CodeLocation] = self._reaching_definitions.all_uses.get_uses(the_def)

            # TODO: We can only replace all these uses with the stack variable if the stack variable isn't
            # TODO: re-assigned of a new value. Perform this check.

            # replace all uses
            for u in all_uses:
                if u == eq.codeloc:
                    # skip the very initial assignment location
                    continue
                old_block = addr2block.get(u.block_addr, None)
                if old_block is None:
                    continue

                # if there is an updated block, use it
                the_block = self.blocks.get(old_block, old_block)
                stmt: Statement = the_block.statements[u.stmt_idx]

                if isinstance(eq.atom0, SimStackVariable):
                    # create the memory loading expression
                    dst = Load(None, StackBaseOffset(None, self.project.arch.bits, eq.atom0.offset), eq.atom0.size,
                               endness=self.project.arch.memory_endness)
                elif isinstance(eq.atom0, Register):
                    dst = eq.atom0
                else:
                    raise RuntimeError("Unsupported atom0 type %s." % type(eq.atom0))

                self._replace_expr_and_update_block(the_block, u.stmt_idx, stmt, the_def, eq.atom1, dst)

    def _fold_call_exprs(self):
        """
        Fold a call expression (statement) into other statements if the return value of the call expression (statement)
        is only used once, and the use site and the call site belongs to the same supernode.

        Example::

            s0 = func();
            if (s0) ...

        after folding, it will be transformed to::

            if (func()) ...
        """

        prop = self.project.analyses.Propagator(func=self.func, func_graph=self.func_graph)
        if not prop.equivalence:
            return

        addr2block: Dict[int, Block] = { }
        for block in self.func_graph.nodes():
            addr2block[block.addr] = block

        for eq in prop.equivalence:
            eq: Equivalence

            # register variable == Call
            if isinstance(eq.atom0, Register):
                if isinstance(eq.atom1, Call):
                    # register variable = Call
                    call = eq.atom1
                else:
                    continue

                # find the definition of this register
                defs = [ d for d in self._reaching_definitions.all_definitions
                         if d.codeloc == eq.codeloc
                         and isinstance(d.atom, atoms.Register)
                         and d.atom.reg_offset == eq.atom0.reg_offset
                         ]
                if not defs or len(defs) > 1:
                    continue
                the_def: Definition = defs[0]

                # find all uses of this definition
                all_uses: Set[CodeLocation] = self._reaching_definitions.all_uses.get_uses(the_def)

                if len(all_uses) != 1:
                    continue
                u = next(iter(all_uses))

                # check if the use and the definition is within the same supernode
                super_node_blocks = self._get_super_node_blocks(addr2block[the_def.codeloc.block_addr])
                if u.block_addr not in set(b.addr for b in super_node_blocks):
                    continue

                # replace all uses
                old_block = addr2block.get(u.block_addr, None)
                if old_block is None:
                    continue

                # if there is an updated block, use that
                the_block = self.blocks.get(old_block, old_block)
                stmt: Statement = the_block.statements[u.stmt_idx]

                if isinstance(eq.atom0, Register):
                    src = eq.atom0
                    dst = call
                else:
                    continue

                replaced = self._replace_expr_and_update_block(the_block, u.stmt_idx, stmt, the_def, src, dst)

                if replaced:
                    # this call has been folded to the use site. we can remove this call.
                    self._calls_to_remove.add(eq.codeloc)

    def _get_super_node_blocks(self, start_node: Block) -> List[Block]:

        lst: List[Block] = [ start_node ]
        while True:
            b = lst[-1]
            successors = list(self.func_graph.successors(b))
            if len(successors) == 0:
                break
            if len(successors) == 1:
                succ = successors[0]
                # check its predecessors
                succ_predecessors = list(self.func_graph.predecessors(succ))
                if len(succ_predecessors) == 1:
                    lst.append(succ)
                else:
                    break
            else:
                # too many successors
                break
        return lst

    def _replace_expr_and_update_block(self, block, stmt_idx, stmt, the_def, src_expr, dst_expr) -> bool:
        replaced, new_stmt = stmt.replace(src_expr, dst_expr)
        if replaced:
            new_block = block.copy()
            new_block.statements = block.statements[::]
            new_block.statements[stmt_idx] = new_stmt
            self.blocks[block] = new_block

            # update the uses
            self._reaching_definitions.all_uses.remove_uses(the_def)
            return True

        return False

    def _remove_dead_assignments(self):

        stmts_to_remove_per_block = defaultdict(set)

        # Find all statements that should be removed

        for def_ in self._reaching_definitions.all_definitions:  # type: Definition
            if def_.dummy:
                continue
            # we do not remove references to global memory regions no matter what
            if isinstance(def_.atom, atoms.MemoryLocation) and isinstance(def_.atom.addr, int):
                continue
            if not self._remove_dead_memdefs and isinstance(def_.atom, atoms.MemoryLocation):
                continue
            uses = self._reaching_definitions.all_uses.get_uses(def_)

            if not uses:
                stmts_to_remove_per_block[def_.codeloc.block_addr].add(def_.codeloc.stmt_idx)

        # Remove the statements
        for old_block in self.func_graph.nodes():

            # if there is an updated block, use it
            block = self.blocks.get(old_block, old_block)

            if not isinstance(block, Block):
                continue

            if block.addr not in stmts_to_remove_per_block:
                continue

            new_statements = [ ]
            stmts_to_remove = stmts_to_remove_per_block[block.addr]

            for idx, stmt in enumerate(block.statements):
                if idx in stmts_to_remove:
                    if isinstance(stmt, (Assignment, Store)):
                        # Skip Assignment and Store statements
                        # if this statement triggers a call, it should not be removed
                        if not self._statement_has_call_exprs(stmt):
                            continue
                    elif isinstance(stmt, Call):
                        codeloc = CodeLocation(block.addr, idx, ins_addr=stmt.ins_addr)
                        if codeloc in self._calls_to_remove:
                            # this call can be removed
                            continue
                        # the return expr is not used. it should not have return expr
                        stmt = stmt.copy()
                        stmt.ret_expr = None
                    else:
                        # Should not happen!
                        raise NotImplementedError()

                new_statements.append(stmt)

            new_block = block.copy()
            new_block.statements = new_statements
            self.blocks[old_block] = new_block

    @staticmethod
    def _statement_has_call_exprs(stmt: Statement) -> bool:

        def _handle_callexpr(expr_idx, expr, stmt_idx, stmt, block):  # pylint:disable=unused-argument
            raise HasCallNotification()

        walker = AILBlockWalker()
        walker.expr_handlers[Call] = _handle_callexpr
        try:
            walker.walk_statement(stmt)
        except HasCallNotification:
            return True

        return False


AnalysesHub.register_default("AILSimplifier", AILSimplifier)




import logging

from ..values import SimSootValue_ArrayRef
from .base import SimSootExpr

l = logging.getLogger('angr.engines.soot.expressions.arrayref')


class SimSootExpr_ArrayRef(SimSootExpr):

    def _execute(self):
        array_base_local = self._translate_value(self.expr.base)
        array_base = self.state.memory.load(array_base_local)
        if array_base is not None:
            # translate idx and check against array bounds
            idx = SimSootValue_ArrayRef.translate_array_index(self.expr.index, self.state)
            SimSootValue_ArrayRef.check_array_bounds(idx, array_base, self.state)
            # load element
            array_ref = SimSootValue_ArrayRef(array_base, idx)
            self.expr = self.state.memory.load(array_ref)
        else:
            l.warning("Trying to access a non existing array! (%r)", self.expr)



from typing import Optional, Iterable, Set, TYPE_CHECKING
import logging

import archinfo

from ...knowledge_plugins.key_definitions import LiveDefinitions
from ...knowledge_plugins.key_definitions.atoms import Atom, GuardUse, Register, MemoryLocation
from ...knowledge_plugins.key_definitions.definition import Definition
from ...knowledge_plugins.key_definitions.environment import Environment
from ...knowledge_plugins.key_definitions.tag import InitialValueTag, ParameterTag, Tag
from ...knowledge_plugins.key_definitions.undefined import UNDEFINED
from ...knowledge_plugins.key_definitions.dataset import DataSet
from ...calling_conventions import SimCC, SimRegArg, SimStackArg
from ...engines.light import SpOffset
from ...code_location import CodeLocation
from .external_codeloc import ExternalCodeLocation
from .heap_allocator import HeapAllocator
from .subject import Subject, SubjectType

if TYPE_CHECKING:
    from .reaching_definitions import ReachingDefinitionsAnalysis


l = logging.getLogger(name=__name__)


class ReachingDefinitionsState:
    """
    Represents the internal state of the ReachingDefinitionsAnalysis.

    It contains a data class LiveDefinitions, which stores both definitions and uses for register, stack, memory, and
    temporary variables, uncovered during the analysis.

    :param subject: The subject being analysed.
    :ivar arch: The architecture targeted by the program.
    :param track_tmps: Only tells whether or not temporary variables should be taken into consideration when
                              representing the state of the analysis.
                              Should be set to true when the analysis has counted uses and definitions for temporary
                              variables, false otherwise.
    :param analysis: The analysis that generated the state represented by this object.
    :param rtoc_value: When the targeted architecture is ppc64, the initial function needs to know the `rtoc_value`.
    :param live_definitions:
    :param canonical_size:
        The sizes (in bytes) that objects with an UNKNOWN_SIZE are treated as for operations where sizes are necessary.
    :param heap_allocator: Mechanism to model the management of heap memory.
    :param environment: Representation of the environment of the analysed program.
    """

    __slots__ = ('arch', '_subject', '_track_tmps', 'analysis', 'current_codeloc', 'codeloc_uses', 'live_definitions',
                 'all_definitions', '_canonical_size', 'heap_allocator', '_environment', )

    def __init__(self, arch: archinfo.Arch, subject: Subject, track_tmps: bool=False,
                 analysis: Optional['ReachingDefinitionsAnalysis']=None, rtoc_value=None,
                 live_definitions=None, canonical_size: int=8, heap_allocator: HeapAllocator=None,
                 environment: Environment=None):

        # handy short-hands
        self.arch = arch
        self._subject = subject
        self._track_tmps = track_tmps
        self.analysis = analysis
        self._canonical_size: int = canonical_size

        if live_definitions is None:
            # the first time this state is created. initialize it
            self.live_definitions = LiveDefinitions(self.arch, track_tmps=self._track_tmps,
                                                    canonical_size=canonical_size)
            self._set_initialization_values(subject, rtoc_value)
        else:
            # this state is a copy from a previous state. skip the initialization
            self.live_definitions = live_definitions

        self.all_definitions: Set[Definition] = set()

        self.heap_allocator = heap_allocator or HeapAllocator(canonical_size)
        self._environment: Environment = environment or Environment()

        self.current_codeloc: Optional[CodeLocation] = None
        self.codeloc_uses: Set[Definition] = set()

    @property
    def tmp_definitions(self): return self.live_definitions.tmp_definitions

    @property
    def tmp_uses(self): return self.live_definitions.tmp_uses

    @property
    def register_uses(self): return self.live_definitions.register_uses

    @property
    def register_definitions(self): return self.live_definitions.register_definitions

    @property
    def stack_definitions(self): return self.live_definitions.stack_definitions

    @property
    def stack_uses(self): return self.live_definitions.stack_uses

    @property
    def heap_definitions(self): return self.live_definitions.heap_definitions

    @property
    def heap_uses(self): return self.live_definitions.heap_uses

    @property
    def memory_uses(self): return self.live_definitions.memory_uses

    @property
    def memory_definitions(self): return self.live_definitions.memory_definitions

    @property
    def uses_by_codeloc(self): return self.live_definitions.uses_by_codeloc

    def get_sp(self) -> int: return self.live_definitions.get_sp()

    @property
    def environment(self): return self._environment

    @property
    def dep_graph(self):
        return self.analysis.dep_graph

    def __repr__(self):
        ctnt = "RDState-%r" % (self.live_definitions)
        return "{%s}" % ctnt

    def _set_initialization_values(self, subject: Subject, rtoc_value: Optional[int]=None):
        if subject.type == SubjectType.Function:
            if isinstance(self.arch, archinfo.arch_ppc64.ArchPPC64) and not rtoc_value:
                raise ValueError('The architecture being ppc64, the parameter `rtoc_value` should be provided.')

            self._initialize_function(
                subject.cc,
                subject.content.addr,
                rtoc_value,
            )
        elif subject.type == SubjectType.CallTrace:
            if isinstance(self.arch, archinfo.arch_ppc64.ArchPPC64) and not rtoc_value:
                raise ValueError('The architecture being ppc64, the parameter `rtoc_value` should be provided.')

            self._initialize_function(
                subject.cc,
                subject.content.current_function_address(),
                rtoc_value,
            )
        elif subject.type == SubjectType.Block:
            pass

        return self

    def _initialize_function(self, cc: SimCC, func_addr: int, rtoc_value: Optional[int]=None):
        # initialize stack pointer
        sp = Register(self.arch.sp_offset, self.arch.bytes)
        sp_def = Definition(sp, ExternalCodeLocation(), DataSet(SpOffset(self.arch.bits, 0), self.arch.bits), tags={InitialValueTag()})
        self.register_definitions.set_object(sp_def.offset, sp_def, sp_def.size)
        if self.arch.name.startswith('MIPS'):
            if func_addr is None:
                l.warning("func_addr must not be None to initialize a function in mips")
            t9 = Register(self.arch.registers['t9'][0], self.arch.bytes)
            t9_def = Definition(t9, ExternalCodeLocation(), DataSet(func_addr, self.arch.bits), tags={InitialValueTag()})
            self.register_definitions.set_object(t9_def.offset,t9_def,t9_def.size)

        if cc is not None and cc.args is not None:
            for arg in cc.args:
                # initialize register parameters
                if isinstance(arg, SimRegArg):
                    # FIXME: implement reg_offset handling in SimRegArg
                    reg_offset = self.arch.registers[arg.reg_name][0]
                    reg = Register(reg_offset, self.arch.bytes)
                    reg_def = Definition(reg, ExternalCodeLocation(), DataSet(UNDEFINED, self.arch.bits), tags={ParameterTag()})
                    self.register_definitions.set_object(reg.reg_offset, reg_def, reg.size)
                # initialize stack parameters
                elif isinstance(arg, SimStackArg):
                    sp_offset = SpOffset(self.arch.bits, arg.stack_offset)
                    ml = MemoryLocation(sp_offset, arg.size)
                    ml_def = Definition(ml, ExternalCodeLocation(), DataSet(UNDEFINED, arg.size * 8), tags={ParameterTag()})
                    self.stack_definitions.set_object(arg.stack_offset, ml_def, ml.size)
                else:
                    raise TypeError('Unsupported parameter type %s.' % type(arg).__name__)

        # architecture dependent initialization
        if self.arch.name.lower().find('ppc64') > -1:
            if rtoc_value is None:
                raise TypeError("rtoc_value must be provided on PPC64.")
            offset, size = self.arch.registers['rtoc']
            rtoc = Register(offset, size)
            rtoc_def = Definition(rtoc, ExternalCodeLocation(), DataSet(rtoc_value, self.arch.bits), tags=InitialValueTag())
            self.register_definitions.set_object(rtoc.reg_offset, rtoc_def, rtoc.size)
        elif self.arch.name.lower().find('mips64') > -1:
            offset, size = self.arch.registers['t9']
            t9 = Register(offset, size)
            t9_def = Definition(t9, ExternalCodeLocation(), DataSet({func_addr}, self.arch.bits), tags=InitialValueTag())
            self.register_definitions.set_object(t9.reg_offset, t9_def, t9.size)

    def copy(self) -> 'ReachingDefinitionsState':
        rd = ReachingDefinitionsState(
            self.arch,
            self._subject,
            track_tmps=self._track_tmps,
            analysis=self.analysis,
            live_definitions=self.live_definitions.copy(),
            canonical_size=self._canonical_size,
            heap_allocator=self.heap_allocator,
            environment=self._environment,
        )

        return rd

    def merge(self, *others):

        state = self.copy()

        for other in others:
            other: 'ReachingDefinitionsState'
            state.live_definitions = state.live_definitions.merge(other.live_definitions)
            state._environment = state.environment.merge(other.environment)

        return state

    def _cycle(self, code_loc: CodeLocation) -> None:
        if code_loc != self.current_codeloc:
            self.current_codeloc = code_loc
            self.codeloc_uses = set()

    def kill_definitions(self, atom: Atom, code_loc: CodeLocation, data: Optional[DataSet]=None, dummy=True, tags: Set[Tag]=None) -> None:
        """
        Overwrite existing definitions w.r.t 'atom' with a dummy definition instance. A dummy definition will not be
        removed during simplification.

        :param atom:
        :param CodeLocation code_loc:
        :param object data:
        :return: None
        """

        if data is None:
            data = DataSet(UNDEFINED, atom.size)

        self.kill_and_add_definition(atom, code_loc, data, dummy=dummy, tags=tags)

    def kill_and_add_definition(self, atom: Atom, code_loc: CodeLocation, data: Optional[DataSet],
                                dummy=False, tags: Set[Tag]=None) -> Optional[Definition]:
        self._cycle(code_loc)

        definition: Optional[Definition]
        definition = self.live_definitions.kill_and_add_definition(atom, code_loc, data, dummy=dummy, tags=tags)

        if definition is not None:
            self.all_definitions.add(definition)

            if self.dep_graph is not None:
                stack_use = set(filter(
                    lambda u: isinstance(u.atom, MemoryLocation) and u.atom.is_on_stack,
                    self.codeloc_uses
                ))

                sp_offset = self.arch.sp_offset
                bp_offset = self.arch.bp_offset

                for used in self.codeloc_uses:
                    # sp is always used as a stack pointer, and we do not track dependencies against stack pointers.
                    # bp is sometimes used as a base pointer. we recognize such cases by checking if there is a use to
                    # the stack variable.
                    #
                    # There are two cases for which it is superfluous to report a dependency on (a use of) stack/base
                    # pointers:
                    # - The `Definition` *uses* a `MemoryLocation` pointing to the stack;
                    # - The `Definition` *is* a `MemoryLocation` pointing to the stack.
                    is_using_spbp_while_memory_address_on_stack_is_used = (
                        isinstance(used.atom, Register) and
                        used.atom.reg_offset in (sp_offset, bp_offset) and
                        len(stack_use) > 0
                    )
                    is_using_spbp_to_define_memory_location_on_stack = (
                        isinstance(definition.atom, MemoryLocation) and
                        definition.atom.is_on_stack and
                        isinstance(used.atom, Register) and
                        used.atom.reg_offset in (sp_offset, bp_offset)
                    )

                    if not (
                        is_using_spbp_while_memory_address_on_stack_is_used or
                        is_using_spbp_to_define_memory_location_on_stack
                    ):
                        # Moderately confusing misnomers. This is an edge from a def to a use, since the
                        # "uses" are actually the definitions that we're using and the "definition" is the
                        # new definition; i.e. The def that the old def is used to construct so this is
                        # really a graph where nodes are defs and edges are uses.
                        self.dep_graph.add_edge(used, definition)
                        self.dep_graph.add_dependencies_for_concrete_pointers_of(
                            used,
                            self.analysis.project.kb.cfgs['CFGFast'],
                            self.analysis.project.loader
                        )

        return definition

    def add_use(self, atom: Atom, code_loc) -> None:
        self._cycle(code_loc)
        self.codeloc_uses.update(self.get_definitions(atom))

        self.live_definitions.add_use(atom, code_loc)

    def add_use_by_def(self, definition: Definition, code_loc: CodeLocation) -> None:
        self._cycle(code_loc)
        self.codeloc_uses.add(definition)

        self.live_definitions.add_use_by_def(definition, code_loc)

    def get_definitions(self, atom: Atom) -> Iterable[Definition]:
        return self.live_definitions.get_definitions(atom)

    def mark_guard(self, code_loc: CodeLocation, data: DataSet, target):
        self._cycle(code_loc)
        atom = GuardUse(target)
        kinda_definition = Definition(atom, code_loc, data)

        if self.dep_graph is not None:
            for used in self.codeloc_uses:
                self.dep_graph.add_edge(used, kinda_definition)




import collections
from itertools import dropwhile
import logging
from typing import Iterator, Optional

from .plugin import SimStatePlugin
from ..errors import AngrError, SimEmptyCallStackError

l = logging.getLogger(name=__name__)

class CallStack(SimStatePlugin):
    """
    Stores the address of the function you're in and the value of SP
    at the VERY BOTTOM of the stack, i.e. points to the return address.
    """
    def __init__(self, call_site_addr=0, func_addr=0, stack_ptr=0, ret_addr=0, jumpkind='Ijk_Call', next_frame: Optional['CallStack'] = None,
                 invoke_return_variable=None):
        super(CallStack, self).__init__()
        self.state = None
        self.call_site_addr = call_site_addr
        self.func_addr = func_addr
        self.stack_ptr = stack_ptr
        self.ret_addr = ret_addr
        self.jumpkind = jumpkind
        self.next = next_frame
        self.invoke_return_variable = invoke_return_variable

        self.block_counter = collections.Counter()
        self.procedure_data = None
        self.locals = {}

    # deprecated as SHIT
    @property
    def call_target(self):
        raise Exception("FIX ME")

    @property
    def return_target(self):
        raise Exception("FIX ME")

    @property
    def stack_pointer(self):
        raise Exception("FIX ME")

    #
    # Public methods
    #

    @SimStatePlugin.memo
    def copy(self, memo, with_tail=True): # pylint: disable=unused-argument,arguments-differ
        n = CallStack(
                call_site_addr=self.call_site_addr,
                func_addr=self.func_addr,
                stack_ptr=self.stack_ptr,
                ret_addr=self.ret_addr,
                jumpkind=self.jumpkind,
                next_frame=self.next if with_tail else None,
                invoke_return_variable=self.invoke_return_variable)
        n.block_counter = collections.Counter(self.block_counter)
        n.procedure_data = self.procedure_data
        n.locals = dict(self.locals)
        return n

    def set_state(self, state):
        super(CallStack, self).set_state(state)
        # make the stack pointer as large as possible as soon as we know how large that actually is
        if self.stack_ptr == 0:
            try:
                bits = state.arch.registers['sp'][1] * state.arch.byte_width
            except KeyError:
                bits = state.arch.bits
            self.stack_ptr = 2**bits - 1

    def merge(self, others, merge_conditions, common_ancestor=None): # pylint: disable=unused-argument
        for o in others:
            if o != self:
                l.error("Trying to merge states with disparate callstacks!")

    def widen(self, others): # pylint: disable=unused-argument
        l.warning("Widening not implemented for callstacks")

    def __iter__(self) -> Iterator['CallStack']:
        """
        Iterate through the callstack, from top to bottom
        (most recent first).
        """
        i = self
        while i is not None:
            yield i
            i = i.next

    def __getitem__(self, k):
        """
        Returns the CallStack at index k, indexing from the top of the stack.
        """
        orig_k = k
        for i in self:
            if k == 0:
                return i
            k -= 1
        raise IndexError(orig_k)

    def __len__(self):
        """
        Get how many frames there are in the current call stack.

        :return: Number of frames
        :rtype: int
        """

        o = 0
        for _ in self:
            o += 1
        return o

    def __repr__(self):
        """
        Get a string representation.

        :return: A printable representation of the CallStack object
        :rtype: str
        """
        return "<CallStack (depth %d)>" % len(self)

    def __str__(self):
        return "Backtrace:\n%s" % "\n".join("Frame %d: %#x => %#x, sp = %#x" % (i, f.call_site_addr, f.func_addr, f.stack_ptr) for i, f in enumerate(self))

    def __eq__(self, other):
        if not isinstance(other, CallStack):
            return False

        if self.func_addr != other.func_addr or self.stack_ptr != other.stack_ptr or self.ret_addr != other.ret_addr:
            return False

        return self.next == other.next

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(tuple((c.func_addr, c.stack_ptr, c.ret_addr) for c in self))

    #
    # Properties
    #

    @property
    def current_function_address(self):
        """
        Address of the current function.

        :return: the address of the function
        :rtype: int
        """

        return self.func_addr

    @current_function_address.setter
    def current_function_address(self, func_addr):
        """
        Set the address of the current function. Note that we must make a copy of the CallStackFrame as CallStackFrame
        is considered to be immutable.

        :param int func_addr: The function address.
        :return: None
        """

        self.func_addr = func_addr

    @property
    def current_stack_pointer(self):
        """
        Get the value of the stack pointer.

        :return: Value of the stack pointer
        :rtype: int
        """
        return self.stack_ptr

    @property
    def current_return_target(self):
        """
        Get the return target.

        :return: The address of return target.
        :rtype: int
        """

        return self.ret_addr

    #
    # Static methods
    #

    @staticmethod
    def stack_suffix_to_string(stack_suffix):
        """
        Convert a stack suffix to a human-readable string representation.
        :param tuple stack_suffix: The stack suffix.
        :return: A string representation
        :rtype: str
        """
        s = "[" + ",".join([("0x%x" % i) if i is not None else "Unspecified" for i in stack_suffix]) + "]"
        return s

    @staticmethod
    def _rfind(lst, item):
        """
        Reverse look-up.

        :param list lst: The list to look up in.
        :param item: The item to look for.
        :return: Offset of the item if found. A ValueError is raised if the item is not in the list.
        :rtype: int
        """

        try:
            return dropwhile(lambda x: lst[x] != item,
                             next(reversed(range(len(lst)))))
        except Exception:
            raise ValueError("%s not in the list" % item)

    @property
    def top(self):
        """
        Returns the element at the top of the callstack without removing it.

        :return: A CallStack.
        """
        return self

    #
    # Public methods
    #

    def push(self, cf):
        """
        Push the frame cf onto the stack. Return the new stack.
        """
        cf.next = self
        if self.state is not None:
            self.state.register_plugin('callstack', cf)
            self.state.history.recent_stack_actions.append(CallStackAction(
                hash(cf), len(cf), 'push', callframe=cf.copy({}, with_tail=False)
            ))

        return cf

    def pop(self):
        """
        Pop the top frame from the stack. Return the new stack.
        """
        if self.next is None:
            raise SimEmptyCallStackError("Cannot pop a frame from an empty call stack.")
        new_list = self.next.copy({})

        if self.state is not None:
            self.state.register_plugin('callstack', new_list)
            self.state.history.recent_stack_actions.append(CallStackAction(
                hash(new_list), len(new_list), 'pop', ret_site_addr=self.ret_addr
            ))

        return new_list

    def call(self, callsite_addr, addr, retn_target=None, stack_pointer=None):
        """
        Push a stack frame into the call stack. This method is called when calling a function in CFG recovery.

        :param int callsite_addr: Address of the call site
        :param int addr: Address of the call target
        :param int or None retn_target: Address of the return target
        :param int stack_pointer: Value of the stack pointer
        :return: None
        """

        frame = CallStack(call_site_addr=callsite_addr, func_addr=addr, ret_addr=retn_target,
                          stack_ptr=stack_pointer)
        return self.push(frame)

    def ret(self, retn_target=None):
        """
        Pop one or many call frames from the stack. This method is called when returning from a function in CFG
        recovery.

        :param int retn_target: The target to return to.
        :return: None
        """

        if retn_target is None:
            return self.pop()

        # We may want to return to several levels up there, not only a
        # single stack frame
        return_target_index = self._find_return_target(retn_target)

        if return_target_index is not None:

            o = self
            while return_target_index >= 0:
                o = o.pop()
                return_target_index -= 1
            return o

        l.warning("Returning to an unexpected address %#x", retn_target)
        return self

            # For Debugging
            # raise Exception()
            # There are cases especially in ARM where return is used as a jump
            # So we don't pop anything out

    def dbg_repr(self):
        """
        Debugging representation of this CallStack object.

        :return: Details of this CalLStack
        :rtype: str
        """

        stack = [ ]
        for i, frame in enumerate(self):
            s = "%d | %s -> %s, returning to %s" % (
                i,
                "None" if frame.call_site_addr is None else "%#x" % frame.call_site_addr,
                "None" if frame.func_addr is None else "%#x" % frame.func_addr,
                "None" if frame.return_target is None else "%#x" % frame.return_target,
            )
            stack.append(s)

        return "\n".join(stack)

    def stack_suffix(self, context_sensitivity_level):
        """
        Generate the stack suffix. A stack suffix can be used as the key to a SimRun in CFG recovery.

        :param int context_sensitivity_level: Level of context sensitivity.
        :return: A tuple of stack suffix.
        :rtype: tuple
        """

        ret = ()

        for frame in self:
            if len(ret) >= context_sensitivity_level*2:
                break
            ret = (frame.call_site_addr, frame.func_addr) + ret

        while len(ret) < context_sensitivity_level*2:
            ret = (None, None) + ret

        return ret

    #
    # Private methods
    #

    def _find_return_target(self, target):
        """
        Check if the return target exists in the stack, and return the index if exists. We always search from the most
        recent call stack frame since the most recent frame has a higher chance to be hit in normal CFG recovery.

        :param int target: Target of the return.
        :return: The index of the object
        :rtype: int
        """

        for i, frame in enumerate(self):
            if frame.ret_addr == target:
                return i
        return None

class CallStackAction(object):
    """
    Used in callstack backtrace, which is a history of callstacks along a path, to record individual actions occurred
    each time the callstack is changed.
    """
    def __init__(self, callstack_hash, callstack_depth, action, callframe=None, ret_site_addr=None):
        self.callstack_hash = callstack_hash
        self.callstack_depth = callstack_depth
        self.action = action

        if action not in ('push', 'pop'):
            raise AngrError('Unsupported action string "%s".' % action)

        self.callframe = callframe
        self.ret_site_addr = ret_site_addr

        if action == 'push' and self.callframe is None:
            raise AngrError('callframe must be specified when action is "push".')

        elif action == 'pop' and self.callframe is not None:
            raise AngrError('callframe must not be specified when action is "pop".')

    def __repr__(self):
        if self.action == 'push':
            return "<CallStackAction push with %s>" % self.callframe
        else: # pop
            return "<CallStackAction pop, ret site %#x>" % self.ret_site_addr


from angr.sim_state import SimState
SimState.register_default('callstack', CallStack)



from . import MemoryMixin

class InspectMixinHigh(MemoryMixin):
    def store(self, addr, data, size=None, condition=None, endness=None, inspect=True, **kwargs):
        if not inspect or not self.state.supports_inspect:
            super().store(addr, data,
                size=size,
                condition=condition,
                endness=endness,
                inspect=inspect,
                **kwargs)
            return

        if self.category == 'reg':
            self.state._inspect('reg_write', BP_BEFORE,
                reg_write_offset=addr,
                reg_write_length=size,
                reg_write_expr=data,
                reg_write_condition=condition,
                reg_write_endness=endness,
            )
            addr = self.state._inspect_getattr('reg_write_offset', addr)
            size = self.state._inspect_getattr('reg_write_length', size)
            data = self.state._inspect_getattr('reg_write_expr', data)
            condition = self.state._inspect_getattr('reg_write_condition', condition)
            endness = self.state._inspect_getattr('reg_write_endness', endness)
        elif self.category == 'mem':
            self.state._inspect('mem_write', BP_BEFORE,
                mem_write_address=addr,
                mem_write_length=size,
                mem_write_expr=data,
                mem_write_condition=condition,
                mem_write_endness=endness,
            )
            addr = self.state._inspect_getattr('mem_write_address', addr)
            size = self.state._inspect_getattr('mem_write_length', size)
            data = self.state._inspect_getattr('mem_write_expr', data)
            condition = self.state._inspect_getattr('mem_write_condition', condition)
            endness = self.state._inspect_getattr('mem_write_endness', endness)

        super().store(addr, data,
            size=size,
            condition=condition,
            endness=endness,
            inspect=inspect,
            **kwargs)

        if self.category == 'reg': self.state._inspect('reg_write', BP_AFTER)
        elif self.category == 'mem': self.state._inspect('mem_write', BP_AFTER)

    def load(self, addr, size=None, condition=None, endness=None, inspect=True, **kwargs):
        if not inspect or not self.state.supports_inspect:
            return super().load(addr,
                size=size,
                condition=condition,
                endness=endness,
                inspect=inspect,
                **kwargs)

        if self.category == 'reg':
            self.state._inspect('reg_read', BP_BEFORE,
                reg_read_offset=addr,
                reg_read_length=size,
                reg_read_condition=condition,
                reg_read_endness=endness)
            addr = self.state._inspect_getattr("reg_read_offset", addr)
            size = self.state._inspect_getattr("reg_read_length", size)
            condition = self.state._inspect_getattr("reg_read_condition", condition)
            endness = self.state._inspect_getattr("reg_read_endness", endness)
        elif self.category == 'mem':
            self.state._inspect('mem_read', BP_BEFORE,
                mem_read_address=addr,
                mem_read_length=size,
                mem_read_condition=condition,
                mem_read_endness=endness)
            addr = self.state._inspect_getattr("mem_read_address", addr)
            size = self.state._inspect_getattr("mem_read_length", size)
            condition = self.state._inspect_getattr("mem_read_condition", condition)
            endness = self.state._inspect_getattr('mem_read_endness', endness)

        r = super().load(addr,
            size=size,
            condition=condition,
            endness=endness,
            inspect=inspect,
            **kwargs)

        if self.category == 'mem':
            self.state._inspect('mem_read', BP_AFTER, mem_read_expr=r)
            r = self.state._inspect_getattr("mem_read_expr", r)

        elif self.category == 'reg':
            self.state._inspect('reg_read', BP_AFTER, reg_read_expr=r)
            r = self.state._inspect_getattr("reg_read_expr", r)

        return r

    def _add_constraints(self, c, add_constraints=True, inspect=True, **kwargs):

        if inspect and self.state.supports_inspect:
            # tracer uses address_concretization_add_constraints
            add_constraints = self.state._inspect_getattr('address_concretization_add_constraints', add_constraints)

        super()._add_constraints(c, add_constraints=add_constraints, inspect=inspect, **kwargs)


from ...state_plugins.inspect import BP_BEFORE, BP_AFTER



from .region_identifier import RegionIdentifier
from .structurer import Structurer
from .structured_codegen import StructuredCodeGenerator
from .clinic import Clinic
from .region_simplifier import RegionSimplifier
from .decompiler import Decompiler
from .decompilation_options import options, options_by_category
from .block_simplifier import BlockSimplifier
from .callsite_maker import CallSiteMaker
from .ail_simplifier import AILSimplifier
from . import optimization_passes




from . import translate_expr
from ..values import translate_value


class SimSootExpr:
    def __init__(self, expr, state):
        self.expr = expr
        self.state = state

    def process(self):
        self._execute()

    def _execute(self):
        raise NotImplementedError()

    def _translate_expr(self, expr):
        expr_ = translate_expr(expr, self.state)
        return expr_

    def _translate_value(self, value):
        value_ = translate_value(value, self.state)
        return value_



import angr

class gethostbyname(angr.SimProcedure):
    def run(self, name):
        malloc = angr.SIM_PROCEDURES['libc']['malloc']
        place = self.inline_call(malloc, 32).ret_expr
        self.state.memory.store(place, self.state.solver.BVS('h_name', 64, key=('api', 'gethostbyname', 'h_name')), endness='Iend_LE')
        self.state.memory.store(place, self.state.solver.BVS('h_aliases', 64, key=('api', 'gethostbyname', 'h_aliases')), endness='Iend_LE')
        self.state.memory.store(place, self.state.solver.BVS('h_addrtype', 64, key=('api', 'gethostbyname', 'h_addrtype')), endness='Iend_LE')
        self.state.memory.store(place, self.state.solver.BVS('h_length', 64, key=('api', 'gethostbyname', 'h_length')), endness='Iend_LE')
        self.state.memory.store(place, self.state.solver.BVS('h_addr_list', 64, key=('api', 'gethostbyname', 'h_addr_list')), endness='Iend_LE')
        return place



from typing import Optional, Dict, Tuple


class CodeLocation:
    """
    Stands for a specific program point by specifying basic block address and statement ID (for IRSBs), or SimProcedure
    name (for SimProcedures).
    """

    __slots__ = ('block_addr', 'stmt_idx', 'sim_procedure', 'ins_addr', 'context', 'info', )

    def __init__(self, block_addr: int, stmt_idx: int, sim_procedure=None, ins_addr: Optional[int]=None,
                 context: Optional[Tuple[int]]=None, **kwargs):
        """
        Constructor.

        :param block_addr:          Address of the block
        :param stmt_idx:            Statement ID. None for SimProcedures
        :param class sim_procedure: The corresponding SimProcedure class.
        :param ins_addr:            The instruction address.
        :param context:             A tuple that represents the context of this CodeLocation in contextful mode, or
                                    None in contextless mode.
        :param kwargs:              Optional arguments, will be stored, but not used in __eq__ or __hash__.
        """

        self.block_addr: int = block_addr
        self.stmt_idx: int = stmt_idx
        self.sim_procedure = sim_procedure
        self.ins_addr: Optional[int] = ins_addr
        self.context: Optional[Tuple[int]] = context

        self.info: Optional[Dict] = None

        self._store_kwargs(**kwargs)

    def __repr__(self):
        if self.block_addr is None:
            return '<%s>' % self.sim_procedure

        if self.stmt_idx is None:
            s = "<%s%#x(-)" % (
                ("%#x " % self.ins_addr) if self.ins_addr else "",
                self.block_addr,
            )
        else:
            s = "<%s%#x[%d]" % (
                ("%#x id=" % self.ins_addr) if self.ins_addr else "",
                self.block_addr,
                self.stmt_idx,
            )

        if self.context is None:
            s += " contextless"
        else:
            s += " context: %r" % (self.context,)

        ss = [ ]
        if self.info:
            for k, v in self.info.items():
                if v != tuple() and v is not None:
                    ss.append("%s=%s" % (k, v))
            if ss:
                s += " with %s" % ", ".join(ss)
        s += ">"

        return s

    @property
    def short_repr(self):
        if self.ins_addr is not None:
            return "%#x" % self.ins_addr
        else:
            return repr(self)

    def __eq__(self, other):
        """
        Check if self is the same as other.
        """
        return type(self) is type(other) and self.block_addr == other.block_addr and \
                self.stmt_idx == other.stmt_idx and self.sim_procedure is other.sim_procedure and \
                self.context == other.context

    def __hash__(self):
        """
        returns the hash value of self.
        """
        return hash((self.block_addr, self.stmt_idx, self.sim_procedure, self.context))

    def _store_kwargs(self, **kwargs):
        if self.info is None:
            self.info = { }
        for k, v in kwargs.items():
            self.info[k] = v



import angr

from cle.backends.externs.simdata.io_file import io_file_data_for_arch

######################################
# fopen
######################################

def mode_to_flag(mode):
    # TODO improve this: handle mode = strings
    if mode[-1] == ord('b'): # lol who uses windows
        mode = mode[:-1]
    elif mode[-1] == ord('t'): # Rarely modes rt or wt are used, but identical to r and w
        mode = mode[:-1]
    mode = mode.replace(b'c', b'').replace(b'e', b'')
    all_modes = {
        b"r"  : angr.storage.file.Flags.O_RDONLY,
        b"r+" : angr.storage.file.Flags.O_RDWR,
        b"w"  : angr.storage.file.Flags.O_WRONLY | angr.storage.file.Flags.O_CREAT,
        b"w+" : angr.storage.file.Flags.O_RDWR | angr.storage.file.Flags.O_CREAT,
        b"a"  : angr.storage.file.Flags.O_WRONLY | angr.storage.file.Flags.O_CREAT | angr.storage.file.Flags.O_APPEND,
        b"a+" : angr.storage.file.Flags.O_RDWR | angr.storage.file.Flags.O_CREAT | angr.storage.file.Flags.O_APPEND
        }
    if mode not in all_modes:
        raise angr.SimProcedureError('unsupported file open mode %s' % mode)

    return all_modes[mode]

class fopen(angr.SimProcedure):
    #pylint:disable=arguments-differ

    def run(self, p_addr, m_addr):
        strlen = angr.SIM_PROCEDURES['libc']['strlen']

        p_strlen = self.inline_call(strlen, p_addr)
        m_strlen = self.inline_call(strlen, m_addr)
        p_expr = self.state.memory.load(p_addr, p_strlen.max_null_index, endness='Iend_BE')
        m_expr = self.state.memory.load(m_addr, m_strlen.max_null_index, endness='Iend_BE')
        path = self.state.solver.eval(p_expr, cast_to=bytes)
        mode = self.state.solver.eval(m_expr, cast_to=bytes)

        # TODO: handle append
        fd = self.state.posix.open(path, mode_to_flag(mode))

        if fd is None:
            # if open failed return NULL
            return 0
        else:
            # Allocate a FILE struct in heap
            malloc = angr.SIM_PROCEDURES['libc']['malloc']
            io_file_data = io_file_data_for_arch(self.state.arch)
            file_struct_ptr = self.inline_call(malloc, io_file_data['size']).ret_expr

            # Write the fd
            fd_bvv = self.state.solver.BVV(fd, 4 * 8) # int
            self.state.memory.store(file_struct_ptr + io_file_data['fd'],
                                    fd_bvv,
                                    endness=self.state.arch.memory_endness)

            return file_struct_ptr



import logging

from .paged_memory_mixin import PagedMemoryMixin
from ....errors import SimSegfaultException

l = logging.getLogger(__name__)

class StackAllocationMixin(PagedMemoryMixin):
    """
    This mixin adds automatic allocation for a stack region based on the stack_end and stack_size parameters.
    """
    # TODO: multiple stacks. this scheme should scale p well
    # TODO tbh this should be handled by an actual fault handler in simos or something
    def __init__(self, stack_end=None, stack_size=None, stack_perms=None, **kwargs):
        super().__init__(**kwargs)
        self._red_pageno = (stack_end - 1) // self.page_size if stack_end is not None else None
        self._remaining_stack = stack_size
        self._stack_perms = stack_perms

    def copy(self, memo):
        o = super().copy(memo)
        o._red_pageno = self._red_pageno
        o._remaining_stack = self._remaining_stack
        o._stack_perms = self._stack_perms
        return o

    def _initialize_page(self, pageno: int, **kwargs):
        if pageno != self._red_pageno:
            return super()._initialize_page(pageno, **kwargs)

        new_red_pageno = (pageno - 1) % ((1 << self.state.arch.bits) // self.page_size)
        if new_red_pageno in self._pages:
            raise SimSegfaultException(pageno * self.page_size, "stack collided with heap")

        if self._remaining_stack is not None and self._remaining_stack < self.page_size:
            raise SimSegfaultException(pageno * self.page_size, "exhausted stack quota")

        self._red_pageno = new_red_pageno
        if self._remaining_stack is not None:
            self._remaining_stack -= self.page_size

        l.debug("Allocating new stack page at %#x", pageno * self.page_size)

        new_page = PagedMemoryMixin._initialize_default_page(self, pageno, permissions=self._stack_perms, **kwargs)
        return new_page




from . import JNISimProcedure
from ...engines.soot.values import SimSootValue_ThisRef
from .method_calls import CallMethodBase

# pylint: disable=arguments-differ,unused-argument

#
# GetObjectClass
#

class GetObjectClass(JNISimProcedure):

    return_ty = 'reference'

    def run(self, ptr_env, obj_):
        obj = self.state.jni_references.lookup(obj_)
        obj_class = self.state.javavm_classloader.get_class(obj.type)
        return self.state.jni_references.create_new_reference(obj_class)

#
# AllocObject
#

class AllocObject(JNISimProcedure):

    return_ty = 'reference'

    def run(self, ptr_env, obj_class_):
        obj_class = self.state.jni_references.lookup(obj_class_)
        # make sure class is initialized
        self.state.javavm_classloader.init_class(obj_class)
        # return object reference
        obj = SimSootValue_ThisRef(heap_alloc_id=self.state.javavm_memory.get_new_uuid(),
                                   type_=obj_class.name)
        return self.state.jni_references.create_new_reference(obj)

#
# NewObject
#

class NewObject(CallMethodBase):

    return_ty = 'reference'
    obj = None
    local_vars = ('obj',)

    def run(self, ptr_env, obj_class_, method_id_):
        # alloc object
        obj_class = self.state.jni_references.lookup(obj_class_)
        self.state.javavm_classloader.init_class(obj_class)
        self.obj = SimSootValue_ThisRef(heap_alloc_id=self.state.javavm_memory.get_new_uuid(),
                                        type_=obj_class.name)
        # call constructor
        method_id = self.state.jni_references.lookup(method_id_)
        self._invoke(method_id, self.obj, dynamic_dispatch=False)

    def return_from_invocation(self, ptr_env, obj_class_, method_id_):
        return self.state.jni_references.create_new_reference(self.obj)

#
# IsInstanceOf
#

class IsInstanceOf(CallMethodBase):

    return_ty = 'boolean'

    def run(self, ptr_env, obj_, target_class_):
        target_class = self.state.jni_references.lookup(target_class_)
        obj = self.state.jni_references.lookup(obj_)
        obj_class = self.state.javavm_classloader.get_class(obj.type)
        # check if target class equals either the object class or one of its superclasses
        class_hierarchy = self.state.javavm_classloader.get_class_hierarchy(obj_class)
        if target_class in class_hierarchy:
            return self.JNI_TRUE
        else:
            return self.JNI_FALSE

#
# IsSameObject
#

class IsSameObject(JNISimProcedure):

    return_ty = 'boolean'

    def run(self, ptr_env, ref1_, ref2_):
        ref1 = self.state.jni_references.lookup(ref1_)
        ref2 = self.state.jni_references.lookup(ref2_)
        if ref1 == ref2:
            return self.JNI_TRUE
        else:
            return self.JNI_FALSE




import claripy
import ailment


INDENT_DELTA = 2


class EmptyBlockNotice(Exception):
    pass


class MultiNode:
    def __init__(self, nodes):

        # delayed import
        from .graph_region import GraphRegion  # pylint:disable=import-outside-toplevel

        self.nodes = [ ]

        for node in nodes:
            if type(node) is MultiNode:
                self.nodes += node.nodes
            elif type(node) is GraphRegion:
                self.nodes += node.nodes
            else:
                self.nodes.append(node)

    def copy(self):
        return MultiNode(self.nodes[::])

    def __repr__(self):

        addrs = [ ]
        s = ""
        for node in self.nodes:
            if hasattr(node, 'addr'):
                addrs.append(node.addr)
            s = ": %#x-%#x" % (min(addrs), max(addrs))

        return "<MultiNode of %d nodes%s>" % (len(self.nodes), s)

    @property
    def addr(self):
        return self.nodes[0].addr


class BaseNode:

    @staticmethod
    def test_empty_node(node):
        # pylint:disable=simplifiable-if-statement
        if type(node) is ailment.Block:
            if not node.statements:
                return True
            # not empty
            return False
        elif type(node) is CodeNode:
            return BaseNode.test_empty_node(node.node)
        # unsupported node type. probably not empty?
        return False

    @staticmethod
    def test_empty_condition_node(cond_node):

        for node in [ cond_node.true_node, cond_node.false_node ]:
            if node is None:
                continue
            if type(node) is CodeNode and BaseNode.test_empty_node(node.node):
                continue
            if BaseNode.test_empty_node(node):
                continue
            return False

        return True


class SequenceNode(BaseNode):
    def __init__(self, nodes=None):
        self.nodes = nodes if nodes is not None else [ ]

    def __repr__(self):
        if self.addr is None:
            return "<SequenceNode, %d nodes>" % len(self.nodes)
        else:
            return "<SequenceNode %#x, %d nodes>" % (self.addr, len(self.nodes))

    @property
    def addr(self):
        if self.nodes:
            return self.nodes[0].addr
        else:
            return None

    def add_node(self, node):
        self.nodes.append(node)

    def insert_node(self, pos, node):
        self.nodes.insert(pos, node)

    def remove_node(self, node):
        self.nodes.remove(node)

    def node_position(self, node):
        return self.nodes.index(node)

    def copy(self):
        return SequenceNode(nodes=self.nodes[::])

    def dbg_repr(self, indent=0):
        s = ""
        for node in self.nodes:
            s += (node.dbg_repr(indent=indent + INDENT_DELTA))
            s += "\n"

        return s


class CodeNode(BaseNode):
    def __init__(self, node, reaching_condition):
        self.node = node
        self.reaching_condition = reaching_condition

    def __repr__(self):
        if self.addr is not None:
            return "<CodeNode %#x>" % self.addr
        else:
            return "<CodeNode %s>" % repr(self.node)

    @property
    def addr(self):
        if hasattr(self.node, 'addr'):
            return self.node.addr
        else:
            return None

    def dbg_repr(self, indent=0):
        indent_str = indent * " "
        s = ""
        if self.reaching_condition is not None and not claripy.is_true(self.reaching_condition):
            s += (indent_str + "if (<block-missing>; %s)\n" +
                 indent_str + "{\n" +
                 indent_str + "  %s\n" +
                 indent_str + "}") % \
                              (self.reaching_condition, self.node)
        else:
            s += indent_str + str(self.node)

        return s

    def copy(self):
        return CodeNode(self.node, self.reaching_condition)


class ConditionNode(BaseNode):
    def __init__(self, addr, reaching_condition, condition, true_node, false_node=None):
        self.addr = addr
        self.reaching_condition = reaching_condition
        self.condition = condition
        self.true_node = true_node
        self.false_node = false_node

    def dbg_repr(self, indent=0):
        indent_str = indent * " "
        s = ""
        s += (indent_str + "if (<block-missing>; %s)\n" +
              indent_str + "{\n" +
              indent_str + "%s\n" +
              indent_str + "}\n") % (
            self.condition, self.true_node.dbg_repr(indent+2),
        )
        if self.false_node is not None:
            s += (indent_str + "else\n" +
              indent_str + "{\n" +
              indent_str + "%s\n" +
              indent_str + "}") % \
             self.false_node.dbg_repr(indent+2)

        return s

    def __repr__(self):
        if self.addr is not None:
            return "<ConditionNode %#x>" % self.addr
        else:
            return "<ConditionNode (%r|%r)>" % (self.true_node, self.false_node)


class LoopNode(BaseNode):
    def __init__(self, sort, condition, sequence_node, addr=None):
        self.sort = sort
        self.condition = condition
        self.sequence_node = sequence_node
        self._addr = addr

    @property
    def addr(self):
        if self._addr is None:
            return self.sequence_node.addr
        else:
            return self._addr


class BreakNode(BaseNode):
    def __init__(self, addr, target):
        self.addr = addr
        self.target = target


class ContinueNode(BaseNode):
    def __init__(self, addr, target):
        self.addr = addr
        self.target = target


class ConditionalBreakNode(BreakNode):
    def __init__(self, addr, condition, target):
        super(ConditionalBreakNode, self).__init__(addr, target)
        self.condition = condition

    def __repr__(self):
        return "<ConditionalBreakNode %#x target:%#x>" % (self.addr, self.target)


class SwitchCaseNode(BaseNode):
    def __init__(self, switch_expr, cases, default_node, addr=None):
        self.switch_expr = switch_expr
        self.cases = cases
        self.default_node = default_node
        self.addr = addr



from typing import TYPE_CHECKING, List, Set, Optional
from abc import ABC, abstractmethod

if TYPE_CHECKING:
    from angr.code_location import CodeLocation
    from angr.analyses.reaching_definitions.dep_graph import DepGraph
    from angr.analyses.reaching_definitions.rd_state import ReachingDefinitionsState


class FunctionHandler(ABC):
    """
    An abstract base class for function handlers.

    To work properly, we expect function handlers to:
      - Be related to a <ReachingDefinitionsAnalysis>;
      - Provide a `handle_local_function` method.
    """

    @abstractmethod
    def hook(self, analysis):
        """
        A <FunctionHandler> needs information about the context in which it is executed.
        A <ReachingDefinitionsAnalysis> would "hook" into a handler by calling: `<FunctionHandler>.hook(self)`.

        :param angr.analyses.ReachingDefinitionsAnalysis analysis: A <ReachingDefinitionsAnalysis> using this <FunctionHandler>.

        :return FunctionHandler:
        """
        raise NotImplementedError()

    @abstractmethod
    def handle_local_function(self, state: 'ReachingDefinitionsState', function_address: int, call_stack: List,
                              maximum_local_call_depth: int, visited_blocks: Set[int], dep_graph: 'DepGraph',
                              src_ins_addr: Optional[int]=None,
                              codeloc: Optional['CodeLocation']=None):
        """
        :param state: The state at the entry of the function, i.e. the function's input state.
        :param function_address: The address of the function to handle.
        :param call_stack:
        :param maximum_local_call_depth:
        :param visited_blocks: A set of the addresses of the previously visited blocks.
        :param dep_graph: A definition-use graph, where nodes represent definitions, and edges represent uses.
        :param codeloc: The code location of the call to the analysed function.

        :return Tuple[Boolean,LiveDefinitions,List<ailment.Block|Block|CodeNode|CFGNode>,DepGraph]:
        """
        raise NotImplementedError()




import logging
from typing import Optional, List

import networkx

from .structurer_nodes import MultiNode


l = logging.getLogger(name=__name__)


class GraphRegion:
    """
    GraphRegion represents a region of nodes.

    :ivar head:             The head of the region.
    :ivar graph:            The region graph.
    :ivar successors:       A set of successors of nodes in the graph. These successors do not belong to the current
                            region.
    :ivar graph_with_successors:    The region graph that includes successor nodes.
    """

    __slots__ = ('head', 'graph', 'successors', 'graph_with_successors', 'cyclic', )

    def __init__(self, head, graph, successors: Optional[set], graph_with_successors, cyclic):
        self.head = head
        self.graph = graph
        self.successors = successors
        # successors inside graph_with_successors should be treated as read-only. when deep-copying GraphRegion objects,
        # successors inside graph_with_successors are *not* deep copied. therefore, you should never modify any
        # successor node in graph_with_successors. to avoid potential programming errors, just treat
        # graph_with_successors as read-only.
        self.graph_with_successors = graph_with_successors
        self.cyclic = cyclic

    def __repr__(self):
        addrs: List[int] = [ ]
        s = ""
        if self.graph is None:
            # only head is available
            return "<GraphRegion %r>" % self.head

        for node in self.graph.nodes():
            if hasattr(node, 'addr'):
                addrs.append(node.addr)
        if addrs:
            s = ": %#x-%#x" % (min(addrs), max(addrs))

        return "<GraphRegion %r of %d nodes%s>" % (self.head, self.graph.number_of_nodes(), s)

    def recursive_copy(self):

        nodes_map = { }
        new_graph = self._recursive_copy(self.graph, nodes_map)

        if self.graph_with_successors is not None:
            successors = set(nodes_map.get(succ, succ) for succ in self.successors)
            # for performance reasons, successors that are only in graph_with_successors are not recursively copied
            new_graph_with_successors = self._recursive_copy(self.graph_with_successors, nodes_map,
                                                             ignored_nodes=successors)
        else:
            new_graph_with_successors = None
            successors = None

        return GraphRegion(nodes_map[self.head], new_graph, successors, new_graph_with_successors, self.cyclic)

    @staticmethod
    def _recursive_copy(old_graph, nodes_map, ignored_nodes=None):
        new_graph = networkx.DiGraph()

        # make copy of each node and add the mapping from old nodes to new nodes into nodes_map
        for node in old_graph.nodes():
            if node in nodes_map:
                new_graph.add_node(nodes_map[node])
            elif ignored_nodes is not None and node in ignored_nodes:
                # do not copy. use the reference instead
                new_graph.add_node(node)
                # drop it into the nodes_map
                nodes_map[node] = node
            else:
                # make recursive copies
                if type(node) is GraphRegion:
                    new_node = node.recursive_copy()
                    nodes_map[node] = new_node
                elif type(node) is MultiNode:
                    new_node = node.copy()
                    nodes_map[node] = new_node
                else:
                    new_node = node
                    nodes_map[node] = new_node
                new_graph.add_node(new_node)

        # add all edges
        for src, dst, edge_data in old_graph.edges(data=True):
            new_graph.add_edge(nodes_map[src], nodes_map[dst], **edge_data)

        return new_graph

    @property
    def addr(self):
        return self.head.addr

    @staticmethod
    def dbg_get_repr(obj, ident=0):
        if type(obj) is GraphRegion:
            s = obj.dbg_print(ident=ident)
        else:
            s = " " * ident + str(obj)

        return s

    def dbg_print(self, ident=0):

        s = self.dbg_get_repr(self.head, ident=ident) + "\n"

        successors = list(self.graph.successors(self.head))
        if len(successors) == 2:
            left_kid, right_kid = successors
            s += " " * ident + "if (...) {\n" + \
                 self.dbg_get_repr(left_kid, ident=ident + 2) + "\n" + \
                 " " * ident + "}\n" + \
                 " " * ident + "else if (...) {\n" + \
                 self.dbg_get_repr(right_kid, ident=ident + 2) + "\n" + \
                 " " * ident + "}"
            # TODO: other nodes
        elif len(successors) == 1:
            s += self.dbg_get_repr(successors[0], ident=ident)

        return s

    def replace_region(self, sub_region, replace_with):

        if sub_region not in self.graph:
            l.error("The sub-region to replace must be in the current region. Note that this method is not recursive.")
            raise Exception()

        if sub_region is self.head:
            self.head = replace_with

        self._replace_node_in_graph(self.graph, sub_region, replace_with)
        if self.graph_with_successors is not None:
            self._replace_node_in_graph(self.graph_with_successors, sub_region, replace_with)

    @staticmethod
    def _replace_node_in_graph(graph, node, replace_with):

        in_edges = list(graph.in_edges(node))
        out_edges = list(graph.out_edges(node))

        graph.remove_node(node)
        graph.add_node(replace_with)

        for src, _ in in_edges:
            if src is node:
                graph.add_edge(replace_with, replace_with)
            else:
                graph.add_edge(src, replace_with)

        for _, dst in out_edges:
            if dst is node:
                graph.add_edge(replace_with, replace_with)
            else:
                graph.add_edge(replace_with, dst)

        assert node not in graph



"""
Classes to structure the different types of <Tag>s that can be attached to <Definition>s.

- Tag
    - FunctionTag
        - ParameterTag
        - LocalVariableTag
        - ReturnValueTag
    - InitialValueTag
"""

class Tag:
    """
    A tag for a Definition that can carry different kinds of metadata.
    """
    def __init__(self, metadata: object=None):
        self.metadata = metadata

    def __repr__(self):
        return "<%s {Metadata: %s}>" % (self.__class__.__name__, self.metadata)

class FunctionTag(Tag):
    """
    A tag for a definition created (or used) in the context of a function.
    """

    def __init__(self, function: int=None, metadata: object=None):
        super(FunctionTag, self).__init__(metadata)
        self.function = function

    def __repr__(self):
        if self.function:
            return '<%s {Function: %#x, Metadata:%s}>' % (self.__class__.__name__, self.function, self.metadata)
        else:
            return super(FunctionTag, self).__repr__()


class SideEffectTag(FunctionTag):
    """
    A tag for a definition created or used as a side-effect of a function.

    Example: The <MemoryLocation> pointed by `rdi` during a `sprintf`.
    """


class ParameterTag(FunctionTag):
    """
    A tag for a definition of a parameter.
    """

class LocalVariableTag(FunctionTag):
    """
    A tag for a definiton of a local variable of a function.
    """

class ReturnValueTag(FunctionTag):
    """
    A tag for a definiton of a return value
    of a function.
    """

class InitialValueTag(Tag):
    """
    A tag for a definiton of an initial value
    """



import angr
import itertools

######################################
# malloc
######################################

malloc_mem_counter = itertools.count()

class malloc(angr.SimProcedure):
    #pylint:disable=arguments-differ
    def run(self, sim_size):
        return self.state.heap._malloc(sim_size)



import logging

from angr.procedures.stubs.format_parser import FormatParser

l = logging.getLogger(name=__name__)

class sscanf(FormatParser):
    #pylint:disable=arguments-differ,unused-argument
    def run(self, data, fmt):
        fmt_str = self._parse(1)
        items = fmt_str.interpret(2, self.arg, addr=data)
        return items




class BlockTerminationNotice(Exception):
    pass

class IncorrectLocationException(Exception):
    pass

class SootMethodNotLoadedException(Exception):
    pass

class SootFieldNotLoadedException(Exception):
    pass




import pyvex

from ..utils import looks_like_sql
from ..knowledge_plugins.xrefs import XRef
from . import Analysis, AnalysesHub


class CodeTags:
    HAS_XOR = 'HAS_XOR'
    HAS_BITSHIFTS = 'HAS_BITSHIFTS'
    HAS_SQL = 'HAS_SQL'
    LARGE_SWITCH = 'LARGE_SWITCH'


class CodeTagging(Analysis):

    def __init__(self, func):
        self._function = func
        self.tags = set()

        self.ANALYSES = [
            self.has_xor,
            self.has_bitshifts,
            self.has_sql,
        ]

        self.analyze()

    def analyze(self):
        for analysis in self.ANALYSES:
            tags = analysis()
            if tags:
                self.tags |= tags

    #
    # Handlers
    #

    def has_xor(self):
        """
        Detects if there is any xor operation in the function.

        :return: Tags
        """

        def _has_xor(expr):
            return isinstance(expr, pyvex.IRExpr.Binop) and expr.op.startswith("Iop_Xor")

        found_xor = False

        for block in self._function.blocks:
            if block.size == 0:
                continue
            for stmt in block.vex.statements:
                if isinstance(stmt, pyvex.IRStmt.Put):
                    found_xor = found_xor or _has_xor(stmt.data)
                elif isinstance(stmt, pyvex.IRStmt.WrTmp):
                    found_xor = found_xor or _has_xor(stmt.data)
            if found_xor:
                break

        if found_xor:
            return { CodeTags.HAS_XOR }
        return None

    def has_bitshifts(self):
        """
        Detects if there is any bitwise operation in the function.

        :return: Tags.
        """

        def _has_bitshifts(expr):
            if isinstance(expr, pyvex.IRExpr.Binop):
                return expr.op.startswith("Iop_Shl") or expr.op.startswith("Iop_Shr") \
                       or expr.op.startswith("Iop_Sar")
            return False

        found_bitops = False

        for block in self._function.blocks:
            if block.size == 0:
                continue
            for stmt in block.vex.statements:
                if isinstance(stmt, pyvex.IRStmt.Put):
                    found_bitops = found_bitops or _has_bitshifts(stmt.data)
                elif isinstance(stmt, pyvex.IRStmt.WrTmp):
                    found_bitops = found_bitops or _has_bitshifts(stmt.data)

            if found_bitops:
                break

        if found_bitops:
            return { CodeTags.HAS_BITSHIFTS }
        return None

    def has_sql(self):
        """
        Detects if there is any reference to strings that look like SQL queries.
        """

        # what strings are the current function referencing?
        for block in self._function.blocks:
            if block.size == 0:
                continue
            for ins_addr in block.instruction_addrs:
                xrefs = self.kb.xrefs.get_xrefs_by_ins_addr(ins_addr)
                for xref in xrefs:
                    xref: XRef
                    if xref.memory_data is not None and xref.memory_data.sort == 'string':
                        if looks_like_sql(xref.memory_data.content.decode("utf-8")):
                            return { CodeTags.HAS_SQL }

        return False

AnalysesHub.register_default('CodeTagging', CodeTagging)




import random
import string

from ..func import Func, TestData


digs = string.digits + string.ascii_letters


class TwoOrThree(object):
    def __eq__(self, other):
        if other == 2:
            return True
        if other == 3:
            return True
        return False

    def __ne__(self, other):
        if other == 2:
            return False
        if other == 3:
            return False
        return True


class ThreeOrFour(object):
    def __eq__(self, other):
        if other == 3:
            return True
        if other == 4:
            return True
        return False

    def __ne__(self, other):
        if other == 3:
            return False
        if other == 4:
            return False
        return True


class FourOrFive(object):
    def __eq__(self, other):
        if other == 4:
            return True
        if other == 5:
            return True
        return False

    def __ne__(self, other):
        if other == 4:
            return False
        if other == 5:
            return False
        return True


class int2str(Func):
    def __init__(self):
        super(int2str, self).__init__()
        self.is_signed = False

    def rand_str(self, length, byte_list=None): #pylint disable=no-self-use
        if byte_list is None:
            return "".join(chr(random.randint(0, 255)) for _ in range(length))
        return "".join(random.choice(byte_list) for _ in range(length))

    def num_args(self):
        return 3

    def args(self): #pylint disable=no-self-use
        return ["buf", "len", "val"]

    def get_name(self):
        if self.is_signed:
            return "int2str"
        return "uint2str"

    def gen_input_output_pair(self):
        num = random.randint(-(2 ** 26), 2 ** 26 - 1)
        if not self.is_signed:
            num = abs(num)
        s = str(num)
        test_input = ["A"*15, 15, num]
        test_output = [s, None, None]
        return_val = None
        max_steps = 10
        test = TestData(test_input, test_output, return_val, max_steps)
        return test

    def pre_test(self, func, runner):
        num = random.randint(-(2 ** 26), 2 ** 26 - 1)
        num = abs(num)
        s = str(num)
        test_input = ["A"*15, 15, num]
        test_output = [s, None, None]
        return_val = None
        max_steps = 10
        test = TestData(test_input, test_output, return_val, max_steps)
        if not runner.test(func, test):
            return False

        num = random.randint(-(2 ** 26), 2 ** 26 - 1)
        num = -abs(num)
        s = str(num)
        test_input = ["A"*15, 15, num]
        test_output = [s, None, None]
        return_val = None
        max_steps = 10
        test = TestData(test_input, test_output, return_val, max_steps)
        if not runner.test(func, test):
            self.is_signed = False
        else:
            self.is_signed = True


        return True


class int2str_v2(Func):
    def __init__(self):
        super(int2str_v2, self).__init__()
        self.is_signed = False

    def rand_str(self, length, byte_list=None): #pylint disable=no-self-use
        if byte_list is None:
            return "".join(chr(random.randint(0, 255)) for _ in range(length))
        return "".join(random.choice(byte_list) for _ in range(length))

    def num_args(self):
        return TwoOrThree()

    def args(self): #pylint disable=no-self-use
        return ["val", "buf", "max"]

    def get_name(self):
        if self.is_signed:
            return "int2str_v2"
        return "uint2str_v2"

    def gen_input_output_pair(self):
        num = random.randint(-(2 ** 26), 2 ** 26 - 1)
        if not self.is_signed:
            num = abs(num)
        s = str(num)
        test_input = [num, "A"*15, 12]
        test_output = [None, s, None]
        return_val = None
        max_steps = 10
        test = TestData(test_input, test_output, return_val, max_steps)
        return test

    def pre_test(self, func, runner):
        num = random.randint(-(2 ** 26), 2 ** 26 - 1)
        num = abs(num)
        s = str(num)
        test_input = [num, "A"*15, 12]
        test_output = [None, s, None]
        return_val = None
        max_steps = 10
        test = TestData(test_input, test_output, return_val, max_steps)
        if not runner.test(func, test):
            return False

        num = random.randint(-(2 ** 26), 2 ** 26 - 1)
        num = -abs(num)
        s = str(num)
        test_input = [num, "A"*15, 12]
        test_output = [None, s, None]
        return_val = None
        max_steps = 10
        test = TestData(test_input, test_output, return_val, max_steps)
        if not runner.test(func, test):
            self.is_signed = False
        else:
            self.is_signed = True

        return True


class int2str_v3(Func):
    def __init__(self):
        super(int2str_v3, self).__init__()
        self.is_signed = False

    def rand_str(self, length, byte_list=None): #pylint disable=no-self-use
        if byte_list is None:
            return "".join(chr(random.randint(0, 255)) for _ in range(length))
        return "".join(random.choice(byte_list) for _ in range(length))

    def num_args(self):
        return TwoOrThree()

    def args(self): #pylint disable=no-self-use
        return ["buf", "val", "max"]

    def get_name(self):
        if self.is_signed:
            return "int2str_v3"
        return "uint2str_v3"

    def gen_input_output_pair(self):
        num = random.randint(-(2 ** 26), 2 ** 26 - 1)
        if not self.is_signed:
            num = abs(num)
        s = str(num)
        test_input = ["A"*15, num, 12]
        test_output = [s, None, None]
        return_val = None
        max_steps = 10
        test = TestData(test_input, test_output, return_val, max_steps)
        return test

    def pre_test(self, func, runner):
        num = random.randint(-(2 ** 26), 2 ** 26 - 1)
        num = abs(num)
        s = str(num)
        test_input = ["A"*15, num, 12]
        test_output = [s, None, None]
        return_val = None
        max_steps = 10
        test = TestData(test_input, test_output, return_val, max_steps)
        if not runner.test(func, test):
            return False

        num = random.randint(-(2 ** 26), 2 ** 26 - 1)
        num = -abs(num)
        s = str(num)
        test_input = [num, "A"*15, 12]
        test_output = [s, None, None]
        return_val = None
        max_steps = 10
        test = TestData(test_input, test_output, return_val, max_steps)
        if not runner.test(func, test):
            self.is_signed = False
        else:
            self.is_signed = True

        return True


class int2str_v4(Func):
    def __init__(self):
        super(int2str_v4, self).__init__()
        self.is_signed = False

    def rand_str(self, length, byte_list=None): #pylint disable=no-self-use
        if byte_list is None:
            return "".join(chr(random.randint(0, 255)) for _ in range(length))
        return "".join(random.choice(byte_list) for _ in range(length))

    def num_args(self):
        return ThreeOrFour()

    def args(self): #pylint disable=no-self-use
        return ["buf", "val", "base"]

    def get_name(self):
        if self.is_signed:
            return "int2str_v4"
        return "uint2str_v4"

    def gen_input_output_pair(self):
        num = random.randint(-(2 ** 26), 2 ** 26 - 1)
        if not self.is_signed:
            num = abs(num)
        s = str(num)
        test_input = ["A"*15, num, 10]
        test_output = [s, None, None]
        return_val = None
        max_steps = 10
        test = TestData(test_input, test_output, return_val, max_steps)
        return test

    def pre_test(self, func, runner):
        num = random.randint(-(2 ** 26), 2 ** 26 - 1)
        num = abs(num)
        s = hex(num).replace("0x", "").replace("L", "").lower()
        s2 = hex(num).replace("0x", "").replace("L", "").upper()
        test_input = ["A"*15, num, 16]
        test_output = [s, None, None]
        return_val = None
        max_steps = 10
        test = TestData(test_input, test_output, return_val, max_steps)
        test_output2 = [s2, None, None]
        test2 = TestData(test_input, test_output2, return_val, max_steps)
        if not runner.test(func, test) and not runner.test(func, test2):
            return False

        num = random.randint(-(2 ** 26), 2 ** 26 - 1)
        num = -abs(num)
        s = str(num)
        test_input = [num, "A"*15, 10]
        test_output = [s, None, None]
        return_val = None
        max_steps = 10
        test = TestData(test_input, test_output, return_val, max_steps)
        if not runner.test(func, test):
            self.is_signed = False
        else:
            self.is_signed = True

        return True



import angr
import logging

l = logging.getLogger(name=__name__)
#l.setLevel("DEBUG")

""" BOOL GetProcessAffinityMask(
  HANDLE     hProcess,
  PDWORD_PTR lpProcessAffinityMask,
  PDWORD_PTR lpSystemAffinityMask
);
"""
class GetProcessAffinityMask(angr.SimProcedure):
    paffinity_mask = None
    saffinity_mask = None

    def run(self, _, lpProcessAffinityMask, lpSystemAffinityMask): #pylint:disable=arguments-differ
        self.fill_symbolic()
        l.info("Setting symbolic memory at %s %s", str(lpProcessAffinityMask), str(lpSystemAffinityMask))

        self.state.mem[lpProcessAffinityMask].dword = self.paffinity_mask
        self.state.mem[lpSystemAffinityMask].dword = self.saffinity_mask

        return 1

    def fill_symbolic(self):
        self.paffinity_mask = self.state.solver.BVS('lpProcessAffinityMask', 32, key=('api', 'lpProcessAffinityMask'))
        self.saffinity_mask = self.state.solver.BVS('lpSystemAffinityMask', 32, key=('api', 'lpSystemAffinityMask'))

    def fill_concrete(self):
        self.paffinity_mask = self.state.solver.BVV(3, 32)
        self.saffinity_mask = self.state.solver.BVV(3, 32)



from typing import Dict, Tuple

import logging
import pickle
import time
import os
import tempfile
import multiprocessing

from .worker import Worker


_l = logging.getLogger(__name__)
_l.setLevel(logging.INFO)


class Server:
    """
    Server implements the analysis server with a series of control interfaces exposed.

    :ivar project:          An instance of angr.Project.
    :ivar str spill_yard:   A directory to store spilled states.
    :ivar str db:           Path of the database that stores information about spilled states.
    :ivar int max_workers:  Maximum number of workers. Each worker starts a new process.
    :ivar int max_states:   Maximum number of active states for each worker.
    :ivar int staging_max:  Maximum number of inactive states that are kept into memory before spilled onto the disk
                            and potentially be picked up by another worker.
    :ivar bool bucketizer:  Use the Bucketizer exploration strategy.
    :ivar _worker_exit_callback:    A method that will be called upon the exit of each worker.
    """
    def __init__(self, project, spill_yard=None, db=None, max_workers=None, max_states=10, staging_max=10,
                 bucketizer=True, recursion_limit=1000, worker_exit_callback=None, techniques=None, add_options=None,
                 remove_options=None):

        self.project = project

        self.spill_yard = spill_yard if spill_yard else tempfile.mkdtemp(suffix="angr_spill_yard")
        if not spill_yard:
            _l.info("Temporary spill yard: %s", self.spill_yard)
        self.db_str = db if db else "sqlite:///" + os.path.join(tempfile.mkdtemp(suffix="angr_server_db"), "db.sqlite3")
        if not db:
            _l.info("Database: %s", self.db_str)

        self.max_workers = max_workers if max_workers is not None else multiprocessing.cpu_count()
        self.max_states = max_states
        self.staging_max = staging_max
        self.bucketizer = bucketizer
        self.techniques = techniques
        self.add_options = add_options
        self.remove_options = remove_options

        self._recursion_limit = recursion_limit

        self._worker_exit_args_lock = None
        self._worker_exit_args: Dict[int,Tuple] = None

        # the following will not be pickled
        self._worker_exit_callback = worker_exit_callback

        self._workers = [ ]
        self._stopped = False
        self._active_workers = multiprocessing.Value('i', lock=True)  # type: multiprocessing.Value

    def __setstate__(self, state):

        self.project = state['project']
        self.spill_yard = state['spill_yard']
        self.db_str = state['db_str']

        self.max_states = state['max_states']
        self.max_workers = state['max_workers']
        self.staging_max = state['staging_max']
        self.bucketizer = state['bucketizer']
        self._worker_exit_args_lock = state['_worker_exit_args_lock']
        self._worker_exit_args = state['_worker_exit_args']
        self._stopped = state['_stopped']
        self._active_workers = state['_active_workers']

    def __getstate__(self):
        s = {
            'project': self.project,
            'spill_yard': self.spill_yard,
            'db_str': self.db_str,
            'max_states': self.max_states,
            'max_workers': self.max_workers,
            'staging_max': self.staging_max,
            'bucketizer': self.bucketizer,
            '_worker_exit_args_lock': self._worker_exit_args_lock,
            '_worker_exit_args': self._worker_exit_args,
            '_stopped': self._stopped,
            '_active_workers': self._active_workers,
        }
        return s

    #
    # Actions
    #

    def inc_active_workers(self):
        with self._active_workers.get_lock():
            self._active_workers.value += 1

    def dec_active_workers(self):
        with self._active_workers.get_lock():
            self._active_workers.value -= 1

    def stop(self):
        self._stopped = True

    #
    # Properties
    #

    @property
    def active_workers(self):
        return self._active_workers.value

    @property
    def stopped(self):
        return self._stopped

    #
    # Callbacks
    #

    def on_worker_exit(self, worker_id, stashes):
        if self._worker_exit_args_lock is not None:
            # callback is enabled
            # we add this check since passing Python objects between processes is definitely not fast
            with self._worker_exit_args_lock:
                self._worker_exit_args[worker_id] = (worker_id, stashes, )

    #
    # Public methods
    #

    def run(self):

        # create workers
        with multiprocessing.Manager() as manager:
            server_state = manager.dict()
            server_state['stopped'] = self.stopped

            if self._worker_exit_callback:
                # Do not initialize the lock if no callback is provided
                self._worker_exit_args_lock = manager.Lock()  # pylint:disable=no-member
                self._worker_exit_args = manager.dict()

            for i in range(self.max_workers):
                _l.info("### Creating worker %d", i)
                worker = Worker(i, self, server_state, recursion_limit=self._recursion_limit,
                                techniques=self.techniques,
                                add_options=self.add_options, remove_options=self.remove_options)
                self._workers.append(worker)

            # start them
            for w in self._workers:
                w.start()

            # should be enough for at least one child process to start
            time.sleep(3)

            i = 0
            while not self.stopped or self.active_workers > 0:
                server_state['stopped'] = self.stopped
                time.sleep(1)

                if self._worker_exit_callback and self._worker_exit_args:
                    with self._worker_exit_args_lock:
                        for _, args in self._worker_exit_args.items():
                            self._worker_exit_callback(*args)

            server_state['stopped'] = self.stopped
            for worker in self._workers:
                # wait for 10 seconds then kill the process
                _l.info("Joining worker %d.", worker.worker_id)
                worker._proc.join(10)
                if worker._proc.is_alive():
                    _l.info("Worker %d is still running. Kill it", worker.worker_id)
                    worker._proc.kill()

            self._workers = [ ]




import logging

from capstone.x86_const import X86_REG_RIP

from .resolver import IndirectJumpResolver

l = logging.getLogger(name=__name__)


class AMD64ElfGotResolver(IndirectJumpResolver):
    def __init__(self, project):
        super().__init__(project, timeless=True)

    def filter(self, cfg, addr, func_addr, block, jumpkind):
        if jumpkind != "Ijk_Call":
            return False

        opnd = self.project.factory.block(addr).capstone.insns[-1].insn.operands[0]
        # Must be of the form: call [rip + 0xABCD]
        if opnd.mem and opnd.mem.disp and opnd.mem.base == X86_REG_RIP and not opnd.mem.index:
            return True
        return False

    def resolve(self, cfg, addr, func_addr, block, jumpkind):
        insn = self.project.factory.block(addr).capstone.insns[-1]
        disp = insn.insn.disp
        slot = disp + insn.address + insn.size
        target = cfg._fast_memory_load_pointer(slot)
        if target is None:
            l.warning("Address %# is not mapped.", slot)
            return False, [ ]

        if not self.project.is_hooked(target):
            return False, [ ]

        dest = self.project.hooked_by(target)
        l.debug("Resolved target to %s", dest.display_name)
        return True, [target]



from typing import Dict, Tuple, Union

from .dataset import dataset_from_datasets, DataSet
from .undefined import Undefined, UNDEFINED
from .unknown_size import UNKNOWN_SIZE


class Environment:
    """
    Represent the environment in which a program runs.
    It's a mapping of variable names, to <DataSet> that should contain possible addresses, or <UNDEFINED>, at
    which their respective values are stored.

    **Note**: The <Environment> object does not store the values associated with variables themselves.
    """
    def __init__(self, environment: Dict[Union[str,Undefined],DataSet]=None):
        self._environment: Dict[Union[str,Undefined],DataSet] = environment or {}

    def get(self, names: DataSet) -> Tuple[DataSet,bool]:
        """
        :param names: Potential values for the name of the environment variable to get the pointers of.
        :return:
            The potential addresses of the values the environment variable can take;
            And a boolean value telling whether all the names were known of the internal representation (i.e. will be
            False if one of the queried variable was not found).
        """
        has_unknown = not all(map(lambda name: name in self._environment.keys(), names))

        def _get(name):
            if not isinstance(name, (str, Undefined)):
                raise TypeError("get(): Expected str, or Undefined, got %s" % type(name).__name__)
            return self._environment.get(name, DataSet({UNDEFINED}, UNKNOWN_SIZE))

        pointers = dataset_from_datasets(list(map(_get, names)))

        return (pointers, has_unknown)

    def set(self, name: Union[str,Undefined], pointers: DataSet):
        """
        :param name: Name of the environment variable to which we will associate the pointers.
        :param pointers: New addresses where the new values of the environment variable are located.
        """
        if not isinstance(name, (str, Undefined)):
            raise TypeError("set(): Expected str, or Undefined, got %s" % type(name).__name__)
        self._environment[name] = pointers

    def __str__(self):
        return "Environment: %s" % self._environment

    def __repr__(self):
        return "Environment: %s" % self._environment

    def __eq__(self, other: 'Environment') -> bool:
        assert isinstance(other, Environment), "Cannot compare Environment with %s" % type(other).__name__
        return self._environment == other._environment

    def merge(self, other: 'Environment'):
        if not isinstance(other, Environment):
            raise TypeError("Cannot merge Environment with %s" % type(other).__name__)

        keys = self._environment.keys() | other._environment.keys()

        def _dataset_from_key(key, environment1, environment2):
            v = environment1.get(key, None)
            w = environment2.get(key, None)
            # Because the key is coming from one of them, they cannot be both `None`.
            if v is None: return w
            if w is None: return v
            return dataset_from_datasets([v, w])

        return Environment(environment=dict(map(
            lambda k: (k, _dataset_from_key(k, self._environment, other._environment)),
            keys
        )))



import angr

from cle.backends.externs.simdata.io_file import io_file_data_for_arch


######################################
# fileno
######################################

class fileno(angr.SimProcedure):
    #pylint:disable=arguments-differ

    def run(self, f):
        # Get FILE struct
        io_file_data = io_file_data_for_arch(self.state.arch)

        # Get the file descriptor from FILE struct
        result = self.state.mem[f + io_file_data['fd']].int.resolved
        return result.sign_extend(self.arch.bits - len(result))

fileno_unlocked = fileno



import angr

class toupper(angr.SimProcedure):
    # pylint: disable=arguments-differ
    def run(self, c):
        return self.state.solver.If(
            self.state.solver.And(c >= 97, c <= 122),  # a - z
            c - 32, c)



from angr.errors import SimSolverError
from ..plugin import SimStatePlugin
from . import SimHeapBase

import logging

l = logging.getLogger(__name__)

class SimHeapBrk(SimHeapBase):
    """
    SimHeapBrk represents a trivial heap implementation based on the Unix `brk` system call. This type of heap stores
    virtually no metadata, so it is up to the user to determine when it is safe to release memory. This also means that
    it does not properly support standard heap operations like `realloc`.

    This heap implementation is a holdover from before any more proper implementations were modelled. At the time,
    various libc (or win32) SimProcedures handled the heap in the same way that this plugin does now. To make future
    heap implementations plug-and-playable, they should implement the necessary logic themselves, and dependent
    SimProcedures should invoke a method by the same name as theirs (prepended with an underscore) upon the heap plugin.
    Depending on the heap implementation, if the method is not supported, an error should be raised.

    Out of consideration for the original way the heap was handled, this plugin implements functionality for all
    relevant SimProcedures (even those that would not normally be supported together in a single heap implementation).

    :ivar heap_location: the address of the top of the heap, bounding the allocations made starting from `heap_base`
    """

    def __init__(self, heap_base=None, heap_size=None):
        super(SimHeapBrk, self).__init__(heap_base, heap_size)
        self.heap_location = self.heap_base

    def allocate(self, sim_size):
        """
        The actual allocation primitive for this heap implementation. Increases the position of the break to allocate
        space. Has no guards against the heap growing too large.

        :param sim_size: a size specifying how much to increase the break pointer by
        :returns: a pointer to the previous break position, above which there is now allocated space
        """
        size = self._conc_alloc_size(sim_size)
        while size % 16 != 0:
            size += 1
        addr = self.state.heap.heap_location
        self.state.heap.heap_location += size
        l.debug("Allocating %d bytes at address %#08x", size, addr)
        return addr

    def release(self, sim_size):
        """
        The memory release primitive for this heap implementation. Decreases the position of the break to deallocate
        space. Guards against releasing beyond the initial heap base.

        :param sim_size: a size specifying how much to decrease the break pointer by (may be symbolic or not)
        """
        requested = self._conc_alloc_size(sim_size)
        used = self.heap_location - self.heap_base
        released = requested if requested <= used else used
        self.heap_location -= released
        l.debug("Releasing %d bytes from the heap (%d bytes were requested to be released)", released, requested)

    def _malloc(self, sim_size):
        return self.allocate(sim_size)

    def _free(self, ptr):  #pylint:disable=unused-argument
        return self.state.solver.Unconstrained('free', self.state.arch.bits)

    def _calloc(self, sim_nmemb, sim_size):
        plugin = self.state.get_plugin('libc')

        if self.state.solver.symbolic(sim_nmemb):
            # TODO: find a better way
            nmemb = self.state.solver.max_int(sim_nmemb)
        else:
            nmemb = self.state.solver.eval(sim_nmemb)

        if self.state.solver.symbolic(sim_size):
            # TODO: find a better way
            size = self.state.solver.max_int(sim_size)
        else:
            size = self.state.solver.eval(sim_size)

        final_size = size * nmemb

        if self.state.solver.symbolic(sim_nmemb) or self.state.solver.symbolic(sim_size):
            if final_size > plugin.max_variable_size:
                final_size = plugin.max_variable_size

        addr = self.state.heap.allocate(final_size)
        v = self.state.solver.BVV(0, final_size * 8)
        self.state.memory.store(addr, v)
        return addr

    def _realloc(self, ptr, size):
        if size.symbolic:
            try:
                size_int = self.state.solver.max(size, extra_constraints=(size < self.state.libc.max_variable_size,))
            except SimSolverError:
                size_int = self.state.solver.min(size)
            self.state.add_constraints(size_int == size)
        else:
            size_int = self.state.solver.eval(size)

        addr = self.state.heap.allocate(size_int)

        if self.state.solver.eval(ptr) != 0:
            v = self.state.memory.load(ptr, size_int)
            self.state.memory.store(addr, v)

        return addr

    @SimStatePlugin.memo
    def copy(self, memo):# pylint: disable=unused-argument
        c = SimHeapBrk(heap_base=self.heap_base, heap_size=self.heap_size)
        c.heap_location = self.heap_location
        c.mmap_base = self.mmap_base
        return c

    def _combine(self, others):
        new_heap_location = max(o.heap_location for o in others)
        if self.heap_location != new_heap_location:
            self.heap_location = new_heap_location
            return True
        else:
            return False

    def merge(self, others, merge_conditions, common_ancestor=None):  #pylint:disable=unused-argument
        return self._combine(others)

    def widen(self, others):
        return self._combine(others)

from angr.sim_state import SimState
SimState.register_default('heap', SimHeapBrk)



from typing import Optional, Set, List, Tuple
import logging

from ...engines.light import SimEngineLight, SpOffset, ArithmeticExpression
from ...errors import SimEngineError
from ...sim_variable import SimVariable, SimStackVariable, SimRegisterVariable, SimMemoryVariable
from ...code_location import CodeLocation
from ..typehoon import typevars, typeconsts

#
# The base engine used in VariableRecoveryFast
#

l = logging.getLogger(name=__name__)


class RichR:
    """
    A rich representation of calculation results.
    """

    __slots__ = ('data', 'variable', 'typevar', 'type_constraints', )

    def __init__(self, data, variable=None, typevar: Optional[typevars.TypeVariable]=None, type_constraints=None):
        self.data = data
        self.variable = variable
        self.typevar = typevar
        self.type_constraints = type_constraints

    @property
    def bits(self):
        if self.data is not None and not isinstance(self.data, (int, float)):
            return self.data.bits
        if self.variable is not None:
            return self.variable.bits
        return None

    def __repr__(self):
        return "R{%r}" % self.data


class SimEngineVRBase(SimEngineLight):
    def __init__(self, project, kb):
        super().__init__()

        self.project = project
        self.kb = kb
        self.processor_state = None
        self.variable_manager = None

    @property
    def func_addr(self):
        if self.state is None:
            return None
        return self.state.function.addr

    def process(self, state, *args, **kwargs):  # pylint:disable=unused-argument

        self.processor_state = state.processor_state
        self.variable_manager = state.variable_manager

        try:
            self._process(state, None, block=kwargs.pop('block', None))
        except SimEngineError as e:
            if kwargs.pop('fail_fast', False) is True:
                raise e

    def _process(self, state, successors, block=None, func_addr=None):  # pylint:disable=unused-argument,arguments-differ
        super()._process(state, successors, block=block)

    #
    # Logic
    #

    def _reference(self, richr: RichR, codeloc: CodeLocation, src=None):
        data = richr.data
        stack_offset = data.offset
        existing_vars: List[Tuple[SimVariable,int]] = self.variable_manager[self.func_addr].find_variables_by_stmt(
            self.block.addr,
            self.stmt_idx,
            'memory')

        # find the correct variable
        variable = None
        for v, offset in existing_vars:
            if offset == stack_offset:
                variable = v
                break

        if variable is None:
            # TODO: how to determine the size for a lea?
            existing_vars = self.state.stack_region.get_variables_by_offset(stack_offset)
            if not existing_vars:
                lea_size = 1
                variable = SimStackVariable(stack_offset, lea_size, base='bp',
                                            ident=self.variable_manager[self.func_addr].next_variable_ident(
                                                'stack'),
                                            region=self.func_addr,
                                            )

                self.variable_manager[self.func_addr].add_variable('stack', stack_offset, variable)
                l.debug('Identified a new stack variable %s at %#x.', variable, self.ins_addr)
            else:
                variable = next(iter(existing_vars))

        self.state.stack_region.add_variable(stack_offset, variable)
        typevar = typevars.TypeVariable() if richr.typevar is None else richr.typevar
        self.state.typevars.add_type_variable(variable, codeloc, typevar)
        base_offset = self.state.stack_region.get_base_addr(stack_offset)
        for var in self.state.stack_region.get_variables_by_offset(base_offset):
            offset_into_var = stack_offset - base_offset
            if offset_into_var == 0: offset_into_var = None
            self.variable_manager[self.func_addr].reference_at(var, offset_into_var, codeloc,
                                                               atom=src)

    def _assign_to_register(self, offset, richr, size, src=None, dst=None):
        """

        :param int offset:
        :param RichR data:
        :param int size:
        :return:
        """

        codeloc: CodeLocation = self._codeloc()
        data = richr.data

        if offset == self.arch.sp_offset:
            if type(data) is SpOffset:
                sp_offset = data.offset
                if isinstance(sp_offset, int):
                    self.processor_state.sp_adjusted = True
                    self.processor_state.sp_adjustment = sp_offset
                    l.debug('Adjusting stack pointer at %#x with offset %+#x.', self.ins_addr, sp_offset)
                elif (isinstance(sp_offset, ArithmeticExpression)
                      and sp_offset.op == ArithmeticExpression.And
                      and isinstance(sp_offset.operands[0], SpOffset)
                      and isinstance(sp_offset.operands[1], int)):
                    l.debug('Masking stack pointer at %#x with mask %#x.', self.ins_addr, sp_offset.operands[1])
                    # ignore masking
                else:
                    l.debug('An unsupported arithmetic expression %r is assigned to stack pointer at %#x. Ignore.',
                            sp_offset,
                            self.ins_addr,
                            )
                    # ignore unsupported arithmetic expressions.
            return

        if offset == self.arch.bp_offset:
            if data is not None:
                self.processor_state.bp = data
            else:
                self.processor_state.bp = None
            return

        if type(data) is SpOffset and isinstance(data.offset, int):
            # lea
            self._reference(richr, codeloc, src=src)
        else:
            pass

        # handle register writes
        existing_vars = self.variable_manager[self.func_addr].find_variables_by_atom(self.block.addr, self.stmt_idx,
                                                                                     dst)
        existing_vars: Set[SimVariable,int]
        if not existing_vars:
            variable = SimRegisterVariable(offset, size,
                                           ident=self.variable_manager[self.func_addr].next_variable_ident(
                                               'register'),
                                           region=self.func_addr
                                           )
            self.variable_manager[self.func_addr].set_variable('register', offset, variable)
        else:
            variable, _ = next(iter(existing_vars))

        self.state.register_region.set_variable(offset, variable)
        self.variable_manager[self.func_addr].write_to(variable, None, codeloc, atom=dst)

        if not self.arch.is_artificial_register(offset, size) and richr.typevar is not None:
            if not self.state.typevars.has_type_variable_for(variable, codeloc):
                # assign a new type variable to it
                typevar = typevars.TypeVariable()
                self.state.typevars.add_type_variable(variable, codeloc, typevar)
                # create constraints
                self.state.add_type_constraint(typevars.Subtype(richr.typevar, typevar))
                self.state.add_type_constraint(typevars.Subtype(typevar, typeconsts.int_type(variable.size * 8)))

    def _store(self, richr_addr: RichR, data, size, stmt=None):  # pylint:disable=unused-argument
        """

        :param RichR addr:
        :param RichR data:
        :param int size:
        :return:
        """

        addr = richr_addr.data

        if type(addr) is SpOffset:
            # Storing data to stack
            stack_offset = addr.offset
            self._store_to_stack(stack_offset, data, size, stmt=stmt)
            return

        if type(addr) is int:
            self._store_to_global(addr, data, size, stmt=stmt)
            return

        if addr is None:
            # storing to a location specified by a pointer whose value cannot be determined at this point
            self._store_to_variable(richr_addr, size, stmt=stmt)

    def _store_to_stack(self, stack_offset, data, size, stmt=None):
        if stmt is None:
            existing_vars = self.variable_manager[self.func_addr].find_variables_by_stmt(self.block.addr,
                                                                                         self.stmt_idx,
                                                                                         'memory'
                                                                                         )
        else:
            existing_vars = self.variable_manager[self.func_addr].find_variables_by_atom(self.block.addr,
                                                                                         self.stmt_idx,
                                                                                         stmt
                                                                                         )
        if not existing_vars:
            variable = SimStackVariable(stack_offset, size, base='bp',
                                        ident=self.variable_manager[self.func_addr].next_variable_ident(
                                            'stack'),
                                        region=self.func_addr,
                                        )
            if isinstance(stack_offset, int):
                self.variable_manager[self.func_addr].set_variable('stack', stack_offset, variable)
                l.debug('Identified a new stack variable %s at %#x.', variable, self.ins_addr)

        else:
            variable, _ = next(iter(existing_vars))

        if isinstance(stack_offset, int):
            self.state.stack_region.set_variable(stack_offset, variable)
            base_offset = self.state.stack_region.get_base_addr(stack_offset)
            codeloc = CodeLocation(self.block.addr, self.stmt_idx, ins_addr=self.ins_addr)
            for var in self.state.stack_region.get_variables_by_offset(stack_offset):
                offset_into_var = stack_offset - base_offset
                if offset_into_var == 0:
                    offset_into_var = None
                self.variable_manager[self.func_addr].write_to(var,
                                                               offset_into_var,
                                                               codeloc,
                                                               atom=stmt,
                                                               )

            # create type constraints
            if data.typevar is not None:
                if not self.state.typevars.has_type_variable_for(variable, codeloc):
                    typevar = typevars.TypeVariable()
                    self.state.typevars.add_type_variable(variable, codeloc, typevar)
                else:
                    typevar = self.state.typevars.get_type_variable(variable, codeloc)
                if typevar is not None:
                    self.state.add_type_constraint(
                        typevars.Subtype(data.typevar, typevar)
                    )
        # TODO: Create a tv_sp.store.<bits>@N <: typevar type constraint for the stack pointer

    def _store_to_global(self, addr: int, data, size, stmt=None):
        variable_manager = self.variable_manager['global']
        if stmt is None:
            existing_vars = variable_manager.find_variables_by_stmt(self.block.addr, self.stmt_idx, 'memory')
        else:
            existing_vars = variable_manager.find_variables_by_atom(self.block.addr, self.stmt_idx, stmt)
        if not existing_vars:
            variable = SimMemoryVariable(addr, size,
                                        ident=variable_manager.next_variable_ident('global'),
                                        )
            variable_manager.set_variable('global', addr, variable)
            l.debug('Identified a new global variable %s at %#x.', variable, self.ins_addr)

        else:
            variable, _ = next(iter(existing_vars))

        self.state.global_region.set_variable(addr, variable)
        codeloc = CodeLocation(self.block.addr, self.stmt_idx, ins_addr=self.ins_addr)
        for var in self.state.global_region.get_variables_by_offset(addr):
            variable_manager.write_to(var, 0, codeloc, atom=stmt)

        # create type constraints
        if data.typevar is not None:
            if not self.state.typevars.has_type_variable_for(variable, codeloc):
                typevar = typevars.TypeVariable()
                self.state.typevars.add_type_variable(variable, codeloc, typevar)
            else:
                typevar = self.state.typevars.get_type_variable(variable, codeloc)
            if typevar is not None:
                self.state.add_type_constraint(
                    typevars.Subtype(data.typevar, typevar)
                )

    def _store_to_variable(self, richr_addr: RichR, size, stmt=None):  # pylint:disable=unused-argument

        addr_variable = richr_addr.variable
        codeloc = self._codeloc()

        # Storing data into a pointer
        if richr_addr.type_constraints:
            for tc in richr_addr.type_constraints:
                self.state.add_type_constraint(tc)

        if richr_addr.typevar is None:
            typevar = typevars.TypeVariable()
        else:
            typevar = richr_addr.typevar

        if typevar is not None:
            if isinstance(typevar, typevars.DerivedTypeVariable) and isinstance(typevar.label, typevars.AddN):
                base_typevar = typevar.type_var
                field_offset = typevar.label.n
            else:
                base_typevar = typevar
                field_offset = 0

            # if addr_variable is not None:
            #     self.variable_manager[self.func_addr].reference_at(addr_variable, field_offset, codeloc, atom=stmt)

            store_typevar = typevars.DerivedTypeVariable(
                typevars.DerivedTypeVariable(base_typevar, typevars.Store()),
                typevars.HasField(size * 8, field_offset)
            )
            if addr_variable is not None:
                self.state.typevars.add_type_variable(addr_variable, codeloc, typevar)
            self.state.add_type_constraint(typevars.Existence(store_typevar))

    def _load(self, richr_addr, size, expr=None):
        """

        :param RichR richr_addr:
        :param size:
        :return:
        """

        addr = richr_addr.data

        if type(addr) is SpOffset:
            # Loading data from stack
            stack_offset = addr.offset

            # split the offset into a concrete offset and a dynamic offset
            # the stack offset may not be a concrete offset
            # for example, SP-0xe0+var_1
            if type(stack_offset) is ArithmeticExpression:
                if type(stack_offset.operands[0]) is int:
                    concrete_offset = stack_offset.operands[0]
                    dynamic_offset = stack_offset.operands[1]
                elif type(stack_offset.operands[1]) is int:
                    concrete_offset = stack_offset.operands[1]
                    dynamic_offset = stack_offset.operands[0]
                else:
                    # cannot determine the concrete offset. give up
                    concrete_offset = None
                    dynamic_offset = stack_offset
            else:
                # type(stack_offset) is int
                concrete_offset = stack_offset
                dynamic_offset = None

            # decide which base variable is being accessed using the concrete offset
            if concrete_offset is not None and concrete_offset not in self.state.stack_region:
                variable = SimStackVariable(concrete_offset, size, base='bp',
                                            ident=self.variable_manager[self.func_addr].next_variable_ident(
                                                'stack'),
                                            region=self.func_addr,
                                            )
                self.state.stack_region.add_variable(concrete_offset, variable)

                self.variable_manager[self.func_addr].add_variable('stack', concrete_offset, variable)

                l.debug('Identified a new stack variable %s at %#x.', variable, self.ins_addr)

            base_offset = self.state.stack_region.get_base_addr(concrete_offset)
            codeloc = CodeLocation(self.block.addr, self.stmt_idx, ins_addr=self.ins_addr)

            all_vars = self.state.stack_region.get_variables_by_offset(base_offset)
            if len(all_vars) > 1:
                # overlapping variables
                l.warning("Reading memory with overlapping variables: %s. Ignoring all but the first one.",
                          all_vars)

            var = next(iter(all_vars))
            # calculate variable_offset
            if dynamic_offset is None:
                offset_into_variable = concrete_offset - base_offset
                if offset_into_variable == 0:
                    offset_into_variable = None
            else:
                if concrete_offset == base_offset:
                    offset_into_variable = dynamic_offset
                else:
                    offset_into_variable = ArithmeticExpression(ArithmeticExpression.Add,
                                                                (dynamic_offset, concrete_offset - base_offset,)
                                                                )
            data = self.variable_manager[self.func_addr].read_from(var,
                                                                   offset_into_variable,
                                                                   codeloc,
                                                                   atom=expr,
                                                                   # overwrite=True
                                                                   )

            # add delayed type constraints
            if var in self.state.delayed_type_constraints:
                for constraint in self.state.delayed_type_constraints[var]:
                    self.state.add_type_constraint(constraint)
                self.state.delayed_type_constraints.pop(var)
            # create type constraints
            if not self.state.typevars.has_type_variable_for(var, codeloc):
                typevar = typevars.TypeVariable()
                self.state.typevars.add_type_variable(var, codeloc, typevar)
            else:
                typevar = self.state.typevars.get_type_variable(var, codeloc)
            # TODO: Create a tv_sp.load.<bits>@N type variable for the stack variable
            #typevar = typevars.DerivedTypeVariable(
            #    typevars.DerivedTypeVariable(typevar, typevars.Load()),
            #    typevars.HasField(size * 8, 0)
            #)

            return RichR(data, variable=var, typevar=typevar)

        # Loading data from a pointer
        if richr_addr.type_constraints:
            for tc in richr_addr.type_constraints:
                self.state.add_type_constraint(tc)

        # parse the loading offset
        offset = 0
        if (isinstance(richr_addr.typevar, typevars.DerivedTypeVariable) and
                isinstance(richr_addr.typevar.label, typevars.AddN)):
            offset = richr_addr.typevar.label.n
            richr_addr_typevar = richr_addr.typevar.type_var  # unpack
        else:
            richr_addr_typevar = richr_addr.typevar

        if richr_addr_typevar is not None:
            # create a type constraint
            typevar = typevars.DerivedTypeVariable(
                typevars.DerivedTypeVariable(richr_addr_typevar, typevars.Load()),
                typevars.HasField(size * 8, offset)
            )
            self.state.add_type_constraint(typevars.Existence(typevar))
            return RichR(None, typevar=typevar)
        else:
            return RichR(None)

    def _read_from_register(self, offset, size, expr=None):
        """

        :param offset:
        :param size:
        :return:
        """

        codeloc = self._codeloc()

        if offset == self.arch.sp_offset:
            # loading from stack pointer
            return RichR(SpOffset(self.arch.bits, self.processor_state.sp_adjustment, is_base=False))
        elif offset == self.arch.bp_offset:
            return RichR(self.processor_state.bp)

        if offset not in self.state.register_region:
            variable = SimRegisterVariable(offset, size,
                                           ident=self.variable_manager[self.func_addr].next_variable_ident(
                                               'register'),
                                           region=self.func_addr,
                                           )
            self.state.register_region.add_variable(offset, variable)
            self.variable_manager[self.func_addr].add_variable('register', offset, variable)

        for var in self.state.register_region.get_variables_by_offset(offset):
            self.variable_manager[self.func_addr].read_from(var, None, codeloc, atom=expr)

        # we accept the precision loss here by only returning the first variable
        var = next(iter(self.state.register_region.get_variables_by_offset(offset)))
        if self.arch.is_artificial_register(offset, size):
            typevar = None
        else:
            # add delayed type constraints
            if var in self.state.delayed_type_constraints:
                for constraint in self.state.delayed_type_constraints[var]:
                    self.state.add_type_constraint(constraint)
                self.state.delayed_type_constraints.pop(var)

            if var not in self.state.typevars:
                typevar = typevars.TypeVariable()
                self.state.typevars.add_type_variable(var, codeloc, typevar)
            else:
                # FIXME: This is an extremely stupid hack. Fix it later.
                # typevar = next(reversed(list(self.state.typevars[var].values())))
                typevar = self.state.typevars[var]

        return RichR(None, variable=var, typevar=typevar)



from typing import Optional, Iterable, Dict, Set
import logging

import archinfo

from collections import defaultdict

from ...engines.light import SpOffset
from ...keyed_region import KeyedRegion
from ...code_location import CodeLocation
from .atoms import Atom, Register, MemoryLocation, Tmp
from .definition import Definition, Tag
from .heap_address import HeapAddress
from .undefined import UNDEFINED
from .uses import Uses
from .dataset import DataSet


l = logging.getLogger(name=__name__)


class LiveDefinitions:
    """
    A LiveDefinitions instance contains definitions and uses for register, stack, memory, and temporary variables,
    uncovered during the analysis.
    """

    __slots__ = ('arch', 'track_tmps', 'register_definitions', 'stack_definitions', 'heap_definitions',
                 'memory_definitions', 'tmp_definitions', 'register_uses', 'stack_uses', 'heap_uses',
                 'memory_uses', 'uses_by_codeloc', 'tmp_uses', '_canonical_size', )

    def __init__(self, arch: archinfo.Arch, track_tmps: bool=False, canonical_size=8):

        self.arch = arch
        self.track_tmps = track_tmps
        self._canonical_size: int = canonical_size

        self.register_definitions = KeyedRegion(canonical_size=self._canonical_size)
        self.stack_definitions = KeyedRegion(canonical_size=self._canonical_size)
        self.memory_definitions = KeyedRegion(canonical_size=self._canonical_size)
        self.heap_definitions = KeyedRegion(canonical_size=self._canonical_size)
        self.tmp_definitions: Dict[int,Set[Definition]] = {}

        self.register_uses = Uses()
        self.stack_uses = Uses()
        self.heap_uses = Uses()
        self.memory_uses = Uses()
        self.uses_by_codeloc: Dict[CodeLocation,Set[Definition]] = defaultdict(set)
        self.tmp_uses: Dict[int,Set[CodeLocation]] = defaultdict(set)

    def __repr__(self):
        ctnt = "LiveDefs, %d regdefs, %d stackdefs, %d heapdefs, %d memdefs" % (
                len(self.register_definitions),
                len(self.stack_definitions),
                len(self.heap_definitions),
                len(self.memory_definitions),
                )
        if self.tmp_definitions:
            ctnt += ", %d tmpdefs" % len(self.tmp_definitions)
        return "<%s>" % ctnt

    def copy(self) -> 'LiveDefinitions':
        rd = LiveDefinitions(self.arch, track_tmps=self.track_tmps, canonical_size=self._canonical_size)

        rd.register_definitions = self.register_definitions.copy()
        rd.stack_definitions = self.stack_definitions.copy()
        rd.heap_definitions = self.heap_definitions.copy()
        rd.memory_definitions = self.memory_definitions.copy()
        rd.tmp_definitions = self.tmp_definitions.copy()
        rd.register_uses = self.register_uses.copy()
        rd.stack_uses = self.stack_uses.copy()
        rd.heap_uses = self.heap_uses.copy()
        rd.memory_uses = self.memory_uses.copy()
        rd.tmp_uses = self.tmp_uses.copy()

        return rd

    def get_sp(self) -> int:
        """
        Return the concrete value contained by the stack pointer.
        """
        sp_definitions = self.register_definitions.get_objects_by_offset(self.arch.sp_offset)
        [first_value, *other_values] = [ d.data.get_first_element() for d in sp_definitions ]

        # If there are several definitions for SP, all values must be the same.
        if len(sp_definitions) > 1:
            [first_value, *other_values] = [ d.data.get_first_element() for d in sp_definitions ]
            all_have_same_value = all(map(lambda v: v == first_value, other_values))
            assert all_have_same_value

        return first_value

    def merge(self, *others):

        state = self.copy()

        for other in others:
            other: LiveDefinitions
            state.register_definitions.merge(other.register_definitions)
            state.stack_definitions.merge(other.stack_definitions)
            state.heap_definitions.merge(other.heap_definitions)
            state.memory_definitions.merge(other.memory_definitions)

            state.register_uses.merge(other.register_uses)
            state.stack_uses.merge(other.stack_uses)
            state.heap_uses.merge(other.heap_uses)
            state.memory_uses.merge(other.memory_uses)

        return state

    def kill_definitions(self, atom: Atom, code_loc: CodeLocation, data: Optional[DataSet]=None, dummy=True, tags: Set[Tag]=None) -> None:
        """
        Overwrite existing definitions w.r.t 'atom' with a dummy definition instance. A dummy definition will not be
        removed during simplification.

        :param atom:
        :param CodeLocation code_loc:
        :return: None
        """

        data = DataSet(UNDEFINED, atom.size)
        self.kill_and_add_definition(atom, code_loc, data, dummy=dummy, tags=tags)

    def kill_and_add_definition(self, atom: Atom, code_loc: CodeLocation, data: Optional[DataSet],
                                dummy=False, tags: Set[Tag]=None) -> Optional[Definition]:
        data = data or DataSet(UNDEFINED, atom.size)
        definition: Definition = Definition(atom, code_loc, data, dummy=dummy, tags=tags)

        # set_object() replaces kill (not implemented) and add (add) in one step
        if isinstance(atom, Register):
            self.register_definitions.set_object(atom.reg_offset, definition, atom.size)
        elif isinstance(atom, MemoryLocation):
            if isinstance(atom.addr, SpOffset):
                self.stack_definitions.set_object(atom.addr.offset, definition, atom.size)
            elif isinstance(atom.addr, HeapAddress):
                self.heap_definitions.set_object(atom.addr.value, definition, atom.size)
            elif isinstance(atom.addr, int):
                self.memory_definitions.set_object(atom.addr, definition, atom.size)
            else:
                return None
        elif isinstance(atom, Tmp):
            if self.track_tmps:
                self.tmp_definitions[atom.tmp_idx] = { definition }
            else:
                self.tmp_definitions[atom.tmp_idx] = self.uses_by_codeloc[code_loc]
                return None
        else:
            raise NotImplementedError()

        return definition

    def add_use(self, atom: Atom, code_loc: CodeLocation) -> None:
        if isinstance(atom, Register):
            self._add_register_use(atom, code_loc)
        elif isinstance(atom, MemoryLocation):
            if isinstance(atom.addr, SpOffset):
                self._add_stack_use(atom, code_loc)
            elif isinstance(atom.addr, HeapAddress):
                self._add_heap_use(atom, code_loc)
            elif isinstance(atom.addr, int):
                self._add_memory_use(atom, code_loc)
            else:
                # ignore RegisterOffset
                pass
        elif isinstance(atom, Tmp):
            self._add_tmp_use(atom, code_loc)
        else:
            raise TypeError("Unsupported atom type %s." % type(atom))

    def add_use_by_def(self, definition: Definition, code_loc: CodeLocation) -> None:
        if isinstance(definition.atom, Register):
            self._add_register_use_by_def(definition, code_loc)
        elif isinstance(definition.atom, MemoryLocation):
            if isinstance(definition.atom.addr, SpOffset):
                self._add_stack_use_by_def(definition, code_loc)
            elif isinstance(definition.atom.addr, HeapAddress):
                self._add_heap_use_by_def(definition, code_loc)
            elif isinstance(definition.atom.addr, MemoryLocation):
                self._add_memory_use_by_def(definition, code_loc)
            else:
                # ignore RegisterOffset
                pass
        elif type(definition.atom) is Tmp:
            self._add_tmp_use_by_def(definition, code_loc)
        else:
            raise TypeError()

    def get_definitions(self, atom: Atom) -> Iterable[Definition]:
        if isinstance(atom, Register):
            return self.register_definitions.get_objects_by_offset(atom.reg_offset)
        elif isinstance(atom, MemoryLocation):
            if isinstance(atom.addr, SpOffset):
                return self.stack_definitions.get_objects_by_offset(atom.addr.offset)
            elif isinstance(atom.addr, HeapAddress):
                return self.heap_definitions.get_objects_by_offset(atom.addr.value)
            elif isinstance(atom.addr, int):
                return self.memory_definitions.get_objects_by_offset(atom.addr)
            else:
                return [ ]
        elif type(atom) is Tmp:
            return self.tmp_definitions[atom.tmp_idx]
        else:
            raise TypeError()

    def get_definitions_from_atoms(self, atoms: Iterable[Atom]) -> Iterable[Definition]:
        result = set()
        for atom in atoms:
            result |= self.get_definitions(atom)
        return result

    #
    # Private methods
    #

    def _add_register_use(self, atom: Register, code_loc: CodeLocation) -> None:
        # get all current definitions
        current_defs: Iterable[Definition] = self.register_definitions.get_objects_by_offset(atom.reg_offset)

        for current_def in current_defs:
            self._add_register_use_by_def(current_def, code_loc)

    def _add_register_use_by_def(self, def_: Definition, code_loc: CodeLocation) -> None:
        self.register_uses.add_use(def_, code_loc)
        self.uses_by_codeloc[code_loc].add(def_)

    def _add_stack_use(self, atom: MemoryLocation, code_loc: CodeLocation) -> None:

        if not isinstance(atom.addr, SpOffset):
            raise TypeError("Atom %r is not a stack location atom." % atom)

        current_defs = self.stack_definitions.get_objects_by_offset(atom.addr.offset)

        for current_def in current_defs:
            self._add_stack_use_by_def(current_def, code_loc)

    def _add_stack_use_by_def(self, def_: Definition, code_loc: CodeLocation) -> None:
        self.stack_uses.add_use(def_, code_loc)
        self.uses_by_codeloc[code_loc].add(def_)

    def _add_heap_use(self, atom: MemoryLocation, code_loc: CodeLocation) -> None:

        if not isinstance(atom.addr, HeapAddress):
            raise TypeError("Atom %r is not a heap location atom." % atom)

        current_defs = self.heap_definitions.get_objects_by_offset(atom.addr.value)

        for current_def in current_defs:
            self._add_heap_use_by_def(current_def, code_loc)

    def _add_heap_use_by_def(self, def_: Definition, code_loc: CodeLocation) -> None:
        self.heap_uses.add_use(def_, code_loc)
        self.uses_by_codeloc[code_loc].add(def_)

    def _add_memory_use(self, atom: MemoryLocation, code_loc: CodeLocation) -> None:

        # get all current definitions
        current_defs: Iterable[Definition] = self.memory_definitions.get_objects_by_offset(atom.addr)

        for current_def in current_defs:
            self._add_memory_use_by_def(current_def, code_loc)

    def _add_memory_use_by_def(self, def_: Definition, code_loc: CodeLocation) -> None:
        self.memory_uses.add_use(def_, code_loc)
        self.uses_by_codeloc[code_loc].add(def_)

    def _add_tmp_use(self, atom: Tmp, code_loc: CodeLocation) -> None:

        if self.track_tmps:
            defs = self.tmp_definitions[atom.tmp_idx]
            for def_ in defs:
                self._add_tmp_use_by_def(def_, code_loc)
        else:
            defs = self.tmp_definitions[atom.tmp_idx]
            for d in defs:
                assert not type(d.atom) is Tmp
                self.add_use_by_def(d, code_loc)

    def _add_tmp_use_by_def(self, def_: Definition, code_loc: CodeLocation) -> None:

        if not isinstance(def_.atom, Tmp):
            raise TypeError("Atom %r is not a Tmp atom." % def_.atom)

        self.tmp_uses[def_.atom.tmp_idx].add(code_loc)
        self.uses_by_codeloc[code_loc].add(def_)
